# Copyright (C) 2009-2012 Chris Ball <cjb@laptop.org>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

"""
Abstract bug repository data storage to easily support multiple backends.
"""

import copy
import os
import pickle
import types

from libbe.error import NotSupported
import libbe.storage
from libbe.util.tree import Tree
from libbe.util import InvalidObject
import libbe.version
from libbe import TESTING

if TESTING == True:
    import doctest
    import os.path
    import sys
    import unittest

    from libbe.util.utility import Dir

class ConnectionError (Exception):
    pass

class InvalidStorageVersion(ConnectionError):
    def __init__(self, active_version, expected_version=None):
        if expected_version == None:
            expected_version = libbe.storage.STORAGE_VERSION
        msg = 'Storage in "%s" not the expected "%s"' \
            % (active_version, expected_version)
        Exception.__init__(self, msg)
        self.active_version = active_version
        self.expected_version = expected_version

class InvalidID (KeyError):
    def __init__(self, id=None, revision=None, msg=None):
        KeyError.__init__(self, id)
        self.msg = msg
        self.id = id
        self.revision = revision
    def __str__(self):
        if self.msg == None:
            return '%s in revision %s' % (self.id, self.revision)
        return self.msg


class InvalidRevision (KeyError):
    pass

class InvalidDirectory (Exception):
    pass

class DirectoryNotEmpty (InvalidDirectory):
    pass

class NotWriteable (NotSupported):
    def __init__(self, msg):
        NotSupported.__init__(self, 'write', msg)

class NotReadable (NotSupported):
    def __init__(self, msg):
        NotSupported.__init__(self, 'read', msg)

class EmptyCommit(Exception):
    def __init__(self):
        Exception.__init__(self, 'No changes to commit')

class _EMPTY (object):
    """Entry has been added but has no user-set value."""
    pass

class Entry (Tree):
    def __init__(self, id, value=_EMPTY, parent=None, directory=False,
                 children=None):
        if children == None:
            Tree.__init__(self)
        else:
            Tree.__init__(self, children)
        self.id = id
        self.value = value
        self.parent = parent
        if self.parent != None:
            if self.parent.directory == False:
                raise InvalidDirectory(
                    'Non-directory %s cannot have children' % self.parent)
            parent.append(self)
        self.directory = directory

    def __str__(self):
        return '<Entry %s: %s>' % (self.id, self.value)

    def __repr__(self):
        return str(self)

    def __cmp__(self, other, local=False):
        if other == None:
            return cmp(1, None)
        if cmp(self.id, other.id) != 0:
            return cmp(self.id, other.id)
        if cmp(self.value, other.value) != 0:
            return cmp(self.value, other.value)
        if local == False:
            if self.parent == None:
                if cmp(self.parent, other.parent) != 0:
                    return cmp(self.parent, other.parent)
            elif self.parent.__cmp__(other.parent, local=True) != 0:
                return self.parent.__cmp__(other.parent, local=True)
            for sc,oc in zip(self, other):
                if sc.__cmp__(oc, local=True) != 0:
                    return sc.__cmp__(oc, local=True)
        return 0

    def _objects_to_ids(self):
        if self.parent != None:
            self.parent = self.parent.id
        for i,c in enumerate(self):
            self[i] = c.id
        return self

    def _ids_to_objects(self, dict):
        if self.parent != None:
            self.parent = dict[self.parent]
        for i,c in enumerate(self):
            self[i] = dict[c]
        return self

class Storage (object):
    """
    This class declares all the methods required by a Storage
    interface.  This implementation just keeps the data in a
    dictionary and uses pickle for persistent storage.
    """
    name = 'Storage'

    def __init__(self, repo='/', encoding='utf-8', options=None):
        self.repo = repo
        self.encoding = encoding
        self.options = options
        self.readable = True  # soft limit (user choice)
        self._readable = True # hard limit (backend choice)
        self.writeable = True  # soft limit (user choice)
        self._writeable = True # hard limit (backend choice)
        self.versioned = False
        self.can_init = True
        self.connected = False

    def __str__(self):
        return '<%s %s %s>' % (self.__class__.__name__, id(self), self.repo)

    def __repr__(self):
        return str(self)

    def version(self):
        """Return a version string for this backend."""
        return libbe.version.version()

    def storage_version(self, revision=None):
        """Return the storage format for this backend."""
        return libbe.storage.STORAGE_VERSION

    def is_readable(self):
        return self.readable and self._readable

    def is_writeable(self):
        return self.writeable and self._writeable

    def init(self):
        """Create a new storage repository."""
        if self.can_init == False:
            raise NotSupported('init',
                               'Cannot initialize this repository format.')
        if self.is_writeable() == False:
            raise NotWriteable('Cannot initialize unwriteable storage.')
        return self._init()

    def _init(self):
        f = open(os.path.join(self.repo, 'repo.pkl'), 'wb')
        root = Entry(id='__ROOT__', directory=True)
        d = {root.id:root}
        pickle.dump(dict((k,v._objects_to_ids()) for k,v in d.items()), f, -1)
        f.close()

    def destroy(self):
        """Remove the storage repository."""
        if self.is_writeable() == False:
            raise NotWriteable('Cannot destroy unwriteable storage.')
        return self._destroy()

    def _destroy(self):
        os.remove(os.path.join(self.repo, 'repo.pkl'))

    def connect(self):
        """Open a connection to the repository."""
        if self.is_readable() == False:
            raise NotReadable('Cannot connect to unreadable storage.')
        self._connect()
        self.connected = True

    def _connect(self):
        try:
            f = open(os.path.join(self.repo, 'repo.pkl'), 'rb')
        except IOError:
            raise ConnectionError(self)
        d = pickle.load(f)
        self._data = dict((k,v._ids_to_objects(d)) for k,v in d.items())
        f.close()

    def disconnect(self):
        """Close the connection to the repository."""
        if self.is_writeable() == False:
            return
        if self.connected == False:
            return
        self._disconnect()
        self.connected = False

    def _disconnect(self):
        f = open(os.path.join(self.repo, 'repo.pkl'), 'wb')
        pickle.dump(dict((k,v._objects_to_ids())
                         for k,v in self._data.items()), f, -1)
        f.close()
        self._data = None

    def add(self, id, *args, **kwargs):
        """Add an entry"""
        if self.is_writeable() == False:
            raise NotWriteable('Cannot add entry to unwriteable storage.')
        if not self.exists(id):
            self._add(id, *args, **kwargs)

    def _add(self, id, parent=None, directory=False):
        if parent == None:
            parent = '__ROOT__'
        p = self._data[parent]
        self._data[id] = Entry(id, parent=p, directory=directory)

    def exists(self, *args, **kwargs):
        """Check an entry's existence"""
        if self.is_readable() == False:
            raise NotReadable('Cannot check entry existence in unreadable storage.')
        return self._exists(*args, **kwargs)

    def _exists(self, id, revision=None):
        return id in self._data

    def remove(self, *args, **kwargs):
        """Remove an entry."""
        if self.is_writeable() == False:
            raise NotSupported('write',
                               'Cannot remove entry from unwriteable storage.')
        self._remove(*args, **kwargs)

    def _remove(self, id):
        if self._data[id].directory == True \
                and len(self.children(id)) > 0:
            raise DirectoryNotEmpty(id)
        e = self._data.pop(id)
        e.parent.remove(e)

    def recursive_remove(self, *args, **kwargs):
        """Remove an entry and all its decendents."""
        if self.is_writeable() == False:
            raise NotSupported('write',
                               'Cannot remove entries from unwriteable storage.')
        self._recursive_remove(*args, **kwargs)

    def _recursive_remove(self, id):
        for entry in reversed(list(self._data[id].traverse())):
            self._remove(entry.id)

    def ancestors(self, *args, **kwargs):
        """Return a list of the specified entry's ancestors' ids."""
        if self.is_readable() == False:
            raise NotReadable('Cannot list parents with unreadable storage.')
        return self._ancestors(*args, **kwargs)

    def _ancestors(self, id=None, revision=None):
        if id == None:
            return []
        ancestors = []
        stack = [id]
        while len(stack) > 0:
            id = stack.pop(0)
            parent = self._data[id].parent
            if parent != None and not parent.id.startswith('__'):
                ancestor = parent.id
                ancestors.append(ancestor)
                stack.append(ancestor)
        return ancestors

    def children(self, *args, **kwargs):
        """Return a list of specified entry's children's ids."""
        if self.is_readable() == False:
            raise NotReadable('Cannot list children with unreadable storage.')
        return self._children(*args, **kwargs)

    def _children(self, id=None, revision=None):
        if id == None:
            id = '__ROOT__'
        return [c.id for c in self._data[id] if not c.id.startswith('__')]

    def get(self, *args, **kwargs):
        """
        Get contents of and entry as they were in a given revision.
        revision==None specifies the current revision.

        If there is no id, return default, unless default is not
        given, in which case raise InvalidID.
        """
        if self.is_readable() == False:
            raise NotReadable('Cannot get entry with unreadable storage.')
        if 'decode' in kwargs:
            decode = kwargs.pop('decode')
        else:
            decode = False
        value = self._get(*args, **kwargs)
        if value != None:
            if decode == True and type(value) != types.UnicodeType:
                return unicode(value, self.encoding)
            elif decode == False and type(value) != types.StringType:
                return value.encode(self.encoding)
        return value

    def _get(self, id, default=InvalidObject, revision=None):
        if id in self._data and self._data[id].value != _EMPTY:
            return self._data[id].value
        elif default == InvalidObject:
            raise InvalidID(id)
        return default

    def set(self, id, value, *args, **kwargs):
        """
        Set the entry contents.
        """
        if self.is_writeable() == False:
            raise NotWriteable('Cannot set entry in unwriteable storage.')
        if type(value) == types.UnicodeType:
            value = value.encode(self.encoding)
        self._set(id, value, *args, **kwargs)

    def _set(self, id, value):
        if id not in self._data:
            raise InvalidID(id)
        if self._data[id].directory == True:
            raise InvalidDirectory(
                'Directory %s cannot have data' % self.parent)
        self._data[id].value = value

class VersionedStorage (Storage):
    """
    This class declares all the methods required by a Storage
    interface that supports versioning.  This implementation just
    keeps the data in a list and uses pickle for persistent
    storage.
    """
    name = 'VersionedStorage'

    def __init__(self, *args, **kwargs):
        Storage.__init__(self, *args, **kwargs)
        self.versioned = True

    def _init(self):
        f = open(os.path.join(self.repo, 'repo.pkl'), 'wb')
        root = Entry(id='__ROOT__', directory=True)
        summary = Entry(id='__COMMIT__SUMMARY__', value='Initial commit')
        body = Entry(id='__COMMIT__BODY__')
        initial_commit = {root.id:root, summary.id:summary, body.id:body}
        d = dict((k,v._objects_to_ids()) for k,v in initial_commit.items())
        pickle.dump([d, copy.deepcopy(d)], f, -1) # [inital tree, working tree]
        f.close()

    def _connect(self):
        try:
            f = open(os.path.join(self.repo, 'repo.pkl'), 'rb')
        except IOError:
            raise ConnectionError(self)
        d = pickle.load(f)
        self._data = [dict((k,v._ids_to_objects(t)) for k,v in t.items())
                      for t in d]
        f.close()

    def _disconnect(self):
        f = open(os.path.join(self.repo, 'repo.pkl'), 'wb')
        pickle.dump([dict((k,v._objects_to_ids())
                          for k,v in t.items()) for t in self._data], f, -1)
        f.close()
        self._data = None

    def _add(self, id, parent=None, directory=False):
        if parent == None:
            parent = '__ROOT__'
        p = self._data[-1][parent]
        self._data[-1][id] = Entry(id, parent=p, directory=directory)

    def _exists(self, id, revision=None):
        if revision == None:
            revision = -1
        else:
            revision = int(revision)
        return id in self._data[revision]

    def _remove(self, id):
        if self._data[-1][id].directory == True \
                and len(self.children(id)) > 0:
            raise DirectoryNotEmpty(id)
        e = self._data[-1].pop(id)
        e.parent.remove(e)

    def _recursive_remove(self, id):
        for entry in reversed(list(self._data[-1][id].traverse())):
            self._remove(entry.id)

    def _ancestors(self, id=None, revision=None):
        if id == None:
            return []
        if revision == None:
            revision = -1
        else:
            revision = int(revision)
        ancestors = []
        stack = [id]
        while len(stack) > 0:
            id = stack.pop(0)
            parent = self._data[revision][id].parent
            if parent != None and not parent.id.startswith('__'):
                ancestor = parent.id
                ancestors.append(ancestor)
                stack.append(ancestor)
        return ancestors

    def _children(self, id=None, revision=None):
        if id == None:
            id = '__ROOT__'
        if revision == None:
            revision = -1
        else:
            revision = int(revision)
        return [c.id for c in self._data[revision][id]
                if not c.id.startswith('__')]

    def _get(self, id, default=InvalidObject, revision=None):
        if revision == None:
            revision = -1
        else:
            revision = int(revision)
        if id in self._data[revision] \
                and self._data[revision][id].value != _EMPTY:
            return self._data[revision][id].value
        elif default == InvalidObject:
            raise InvalidID(id)
        return default

    def _set(self, id, value):
        if id not in self._data[-1]:
            raise InvalidID(id)
        self._data[-1][id].value = value

    def commit(self, *args, **kwargs):
        """
        Commit the current repository, with a commit message string
        summary and body.  Return the name of the new revision.

        If allow_empty == False (the default), raise EmptyCommit if
        there are no changes to commit.
        """
        if self.is_writeable() == False:
            raise NotWriteable('Cannot commit to unwriteable storage.')
        return self._commit(*args, **kwargs)

    def _commit(self, summary, body=None, allow_empty=False):
        if self._data[-1] == self._data[-2] and allow_empty == False:
            raise EmptyCommit
        self._data[-1]["__COMMIT__SUMMARY__"].value = summary
        self._data[-1]["__COMMIT__BODY__"].value = body
        rev = str(len(self._data)-1)
        self._data.append(copy.deepcopy(self._data[-1]))
        return rev

    def revision_id(self, index=None):
        """
        Return the name of the <index>th revision.  The choice of
        which branch to follow when crossing branches/merges is not
        defined.  Revision indices start at 1; ID 0 is the blank
        repository.

        Return None if index==None.

        If the specified revision does not exist, raise InvalidRevision.
        """
        if index == None:
            return None
        try:
            if int(index) != index:
                raise InvalidRevision(index)
        except ValueError:
            raise InvalidRevision(index)
        L = len(self._data) - 1  # -1 b/c of initial commit
        if index >= -L and index <= L:
            return str(index % L)
        raise InvalidRevision(i)

    def changed(self, revision):
        """Return a tuple of lists of ids `(new, modified, removed)` from the
        specified revision to the current situation.
        """
        new = []
        modified = []
        removed = []
        for id,value in self._data[int(revision)].items():
            if id.startswith('__'):
                continue
            if not id in self._data[-1]:
                removed.append(id)
            elif value.value != self._data[-1][id].value:
                modified.append(id)
        for id in self._data[-1]:
            if not id in self._data[int(revision)]:
                new.append(id)
        return (new, modified, removed)


if TESTING == True:
    class StorageTestCase (unittest.TestCase):
        """Test cases for Storage class."""

        Class = Storage

        def __init__(self, *args, **kwargs):
            super(StorageTestCase, self).__init__(*args, **kwargs)
            self.dirname = None

        # this class will be the basis of tests for several classes,
        # so make sure we print the name of the class we're dealing with.
        def _classname(self):
            version = '?'
            try:
                if hasattr(self, 's'):
                    version = self.s.version()
            except:
                pass
            return '%s:%s' % (self.Class.__name__, version)

        def fail(self, msg=None):
            """Fail immediately, with the given message."""
            raise self.failureException(
                '({0}) {1}'.format(self._classname(), msg))

        def failIf(self, expr, msg=None):
            "Fail the test if the expression is true."
            if expr:
                raise self.failureException(
                    '({0}) {1}'.format(self._classname(), msg))

        def failUnless(self, expr, msg=None):
            """Fail the test unless the expression is true."""
            if not expr:
                raise self.failureException(
                    '({0}) {1}'.format(self._classname(), msg))

        def setUp(self):
            """Set up test fixtures for Storage test case."""
            super(StorageTestCase, self).setUp()
            self.dir = Dir()
            self.dirname = self.dir.path
            self.s = self.Class(repo=self.dirname)
            self.assert_failed_connect()
            self.s.init()
            self.s.connect()

        def tearDown(self):
            super(StorageTestCase, self).tearDown()
            self.s.disconnect()
            self.s.destroy()
            self.assert_failed_connect()
            self.dir.cleanup()

        def assert_failed_connect(self):
            try:
                self.s.connect()
                self.fail(
                    "Connected to %(name)s repository before initialising"
                    % vars(self.Class))
            except ConnectionError:
                pass

    class Storage_init_TestCase (StorageTestCase):
        """Test cases for Storage.init method."""

        def test_connect_should_succeed_after_init(self):
            """Should connect after initialization."""
            self.s.connect()

    class Storage_connect_disconnect_TestCase (StorageTestCase):
        """Test cases for Storage.connect and .disconnect methods."""

        def test_multiple_disconnects(self):
            """Should be able to call .disconnect multiple times."""
            self.s.disconnect()
            self.s.disconnect()

    class Storage_add_remove_TestCase (StorageTestCase):
        """Test cases for Storage.add, .remove, and .recursive_remove methods."""

        def test_initially_empty(self):
            """New repository should be empty."""
            self.failUnless(len(self.s.children()) == 0, self.s.children())

        def test_add_identical_rooted(self):
            """Adding entries with the same ID should not increase the number of children.
            """
            for i in range(10):
                self.s.add('some id', directory=False)
                s = sorted(self.s.children())
                self.failUnless(s == ['some id'], s)

        def test_add_rooted(self):
            """Adding entries should increase the number of children (rooted).
            """
            ids = []
            for i in range(10):
                ids.append(str(i))
                self.s.add(ids[-1], directory=(i % 2 == 0))
                s = sorted(self.s.children())
                self.failUnless(s == ids, '\n  %s\n  !=\n  %s' % (s, ids))

        def test_add_nonrooted(self):
            """Adding entries should increase the number of children (nonrooted).
            """
            self.s.add('parent', directory=True)
            ids = []
            for i in range(10):
                ids.append(str(i))
                self.s.add(ids[-1], 'parent', directory=(i % 2 == 0))
                s = sorted(self.s.children('parent'))
                self.failUnless(s == ids, '\n  %s\n  !=\n  %s' % (s, ids))
                s = self.s.children()
                self.failUnless(s == ['parent'], s)

        def test_ancestors(self):
            """Check ancestors lists.
            """
            self.s.add('parent', directory=True)
            for i in range(10):
                i_id = str(i)
                self.s.add(i_id, 'parent', directory=True)
                for j in range(10): # add some grandkids
                    j_id = str(20*(i+1)+j)
                    self.s.add(j_id, i_id, directory=(i%2 == 0))
                    ancestors = sorted(self.s.ancestors(j_id))
                    self.failUnless(ancestors == [i_id, 'parent'],
                        'Unexpected ancestors for %s/%s, "%s"'
                        % (i_id, j_id, ancestors))

        def test_children(self):
            """Non-UUID ids should be returned as such.
            """
            self.s.add('parent', directory=True)
            ids = []
            for i in range(10):
                ids.append('parent/%s' % str(i))
                self.s.add(ids[-1], 'parent', directory=(i % 2 == 0))
                s = sorted(self.s.children('parent'))
                self.failUnless(s == ids, '\n  %s\n  !=\n  %s' % (s, ids))

        def test_grandchildren(self):
            """Grandchildren should not be returned as children.
            """
            self.s.add('parent', directory=True)
            ids = []
            for i in range(5):
                child = 'parent/%s' % str(i)
                directory = (i % 2 == 0)
                ids.append(child)
                self.s.add(child, 'parent', directory=directory)
                if directory:
                    for j in range(3):
                        grandchild = '%s/%s' % (child, str(j))
                        directory = (j % 2 == 0)
                        self.s.add(grandchild, child, directory=directory)
                s = sorted(self.s.children('parent'))
                self.failUnless(s == ids, '\n  %s\n  !=\n  %s' % (s, ids))

        def test_add_invalid_directory(self):
            """Should not be able to add children to non-directories.
            """
            self.s.add('parent', directory=False)
            try:
                self.s.add('child', 'parent', directory=False)
                self.fail(
                    '%s.add() succeeded instead of raising InvalidDirectory'
                    % (vars(self.Class)['name']))
            except InvalidDirectory:
                pass
            try:
                self.s.add('child', 'parent', directory=True)
                self.fail(
                    '%s.add() succeeded instead of raising InvalidDirectory'
                    % (vars(self.Class)['name']))
            except InvalidDirectory:
                pass
            self.failUnless(len(self.s.children('parent')) == 0,
                            self.s.children('parent'))

        def test_remove_rooted(self):
            """Removing entries should decrease the number of children (rooted).
            """
            ids = []
            for i in range(10):
                ids.append(str(i))
                self.s.add(ids[-1], directory=(i % 2 == 0))
            for i in range(10):
                self.s.remove(ids.pop())
                s = sorted(self.s.children())
                self.failUnless(s == ids, '\n  %s\n  !=\n  %s' % (s, ids))

        def test_remove_nonrooted(self):
            """Removing entries should decrease the number of children (nonrooted).
            """
            self.s.add('parent', directory=True)
            ids = []
            for i in range(10):
                ids.append(str(i))
                self.s.add(ids[-1], 'parent', directory=False)#(i % 2 == 0))
            for i in range(10):
                self.s.remove(ids.pop())
                s = sorted(self.s.children('parent'))
                self.failUnless(s == ids, '\n  %s\n  !=\n  %s' % (s, ids))
                if len(s) > 0:
                    s = self.s.children()
                    self.failUnless(s == ['parent'], s)

        def test_remove_directory_not_empty(self):
            """Removing a non-empty directory entry should raise exception.
            """
            self.s.add('parent', directory=True)
            ids = []
            for i in range(10):
                ids.append(str(i))
                self.s.add(ids[-1], 'parent', directory=(i % 2 == 0))
            self.s.remove(ids.pop()) # empty directory removal succeeds
            try:
                self.s.remove('parent') # empty directory removal succeeds
                self.fail(
                    "%s.remove() didn't raise DirectoryNotEmpty"
                    % (vars(self.Class)['name']))
            except DirectoryNotEmpty:
                pass

        def test_recursive_remove(self):
            """Recursive remove should empty the tree."""
            self.s.add('parent', directory=True)
            ids = []
            for i in range(10):
                ids.append(str(i))
                self.s.add(ids[-1], 'parent', directory=True)
                for j in range(10): # add some grandkids
                    self.s.add(str(20*(i+1)+j), ids[-1], directory=(i%2 == 0))
            self.s.recursive_remove('parent')
            s = sorted(self.s.children())
            self.failUnless(s == [], s)

    class Storage_get_set_TestCase (StorageTestCase):
        """Test cases for Storage.get and .set methods."""

        id = 'unlikely id'
        val = 'unlikely value'

        def test_get_default(self):
            """Get should return specified default if id not in Storage.
            """
            ret = self.s.get(self.id, default=self.val)
            self.failUnless(ret == self.val,
                    "%s.get() returned %s not %s"
                    % (vars(self.Class)['name'], ret, self.val))

        def test_get_default_exception(self):
            """Get should raise exception if id not in Storage and no default.
            """
            try:
                ret = self.s.get(self.id)
                self.fail(
                    "%s.get() returned %s instead of raising InvalidID"
                    % (vars(self.Class)['name'], ret))
            except InvalidID:
                pass

        def test_get_initial_value(self):
            """Data value should be default before any value has been set.
            """
            self.s.add(self.id, directory=False)
            val = 'UNLIKELY DEFAULT'
            ret = self.s.get(self.id, default=val)
            self.failUnless(ret == val,
                    "%s.get() returned %s not %s"
                    % (vars(self.Class)['name'], ret, val))

        def test_set_exception(self):
            """Set should raise exception if id not in Storage.
            """
            try:
                self.s.set(self.id, self.val)
                self.fail(
                    "%(name)s.set() did not raise InvalidID"
                    % vars(self.Class))
            except InvalidID:
                pass

        def test_set(self):
            """Set should define the value returned by get.
            """
            self.s.add(self.id, directory=False)
            self.s.set(self.id, self.val)
            ret = self.s.get(self.id)
            self.failUnless(ret == self.val,
                    "%s.get() returned %s not %s"
                    % (vars(self.Class)['name'], ret, self.val))

        def test_unicode_set(self):
            """Set should define the value returned by get.
            """
            val = u'Fran\xe7ois'
            self.s.add(self.id, directory=False)
            self.s.set(self.id, val)
            ret = self.s.get(self.id, decode=True)
            self.failUnless(type(ret) == types.UnicodeType,
                    "%s.get() returned %s not UnicodeType"
                    % (vars(self.Class)['name'], type(ret)))
            self.failUnless(ret == val,
                    "%s.get() returned %s not %s"
                    % (vars(self.Class)['name'], ret, self.val))
            ret = self.s.get(self.id)
            self.failUnless(type(ret) == types.StringType,
                    "%s.get() returned %s not StringType"
                    % (vars(self.Class)['name'], type(ret)))
            s = unicode(ret, self.s.encoding)
            self.failUnless(s == val,
                    "%s.get() returned %s not %s"
                    % (vars(self.Class)['name'], s, self.val))


    class Storage_persistence_TestCase (StorageTestCase):
        """Test cases for Storage.disconnect and .connect methods."""

        id = 'unlikely id'
        val = 'unlikely value'

        def test_get_set_persistence(self):
            """Set should define the value returned by get after reconnect.
            """
            self.s.add(self.id, directory=False)
            self.s.set(self.id, self.val)
            self.s.disconnect()
            self.s.connect()
            ret = self.s.get(self.id)
            self.failUnless(ret == self.val,
                    "%s.get() returned %s not %s"
                    % (vars(self.Class)['name'], ret, self.val))

        def test_empty_get_set_persistence(self):
            """After empty set, get may return either an empty string or default.
            """
            self.s.add(self.id, directory=False)
            self.s.set(self.id, '')
            self.s.disconnect()
            self.s.connect()
            default = 'UNLIKELY DEFAULT'
            ret = self.s.get(self.id, default=default)
            self.failUnless(ret in ['', default],
                    "%s.get() returned %s not in %s"
                    % (vars(self.Class)['name'], ret, ['', default]))

        def test_add_nonrooted_persistence(self):
            """Adding entries should increase the number of children after reconnect.
            """
            self.s.add('parent', directory=True)
            ids = []
            for i in range(10):
                ids.append(str(i))
                self.s.add(ids[-1], 'parent', directory=(i % 2 == 0))
            self.s.disconnect()
            self.s.connect()
            s = sorted(self.s.children('parent'))
            self.failUnless(s == ids, '\n  %s\n  !=\n  %s' % (s, ids))
            s = self.s.children()
            self.failUnless(s == ['parent'], s)

    class VersionedStorageTestCase (StorageTestCase):
        """Test cases for VersionedStorage methods."""

        Class = VersionedStorage

    class VersionedStorage_commit_TestCase (VersionedStorageTestCase):
        """Test cases for VersionedStorage.commit and revision_ids methods."""

        id = 'unlikely id'
        val = 'Some value'
        commit_msg = 'Committing something interesting'
        commit_body = 'Some\nlonger\ndescription\n'

        def _setup_for_empty_commit(self):
            """
            Initialization might add some files to version control, so
            commit those first, before testing the empty commit
            functionality.
            """
            try:
                self.s.commit('Added initialization files')
            except EmptyCommit:
                pass

        def test_revision_id_exception(self):
            """Invalid revision id should raise InvalidRevision.
            """
            try:
                rev = self.s.revision_id('highly unlikely revision id')
                self.fail(
                    "%s.revision_id() didn't raise InvalidRevision, returned %s."
                    % (vars(self.Class)['name'], rev))
            except InvalidRevision:
                pass

        def test_empty_commit_raises_exception(self):
            """Empty commit should raise exception.
            """
            self._setup_for_empty_commit()
            try:
                self.s.commit(self.commit_msg, self.commit_body)
                self.fail(
                    "Empty %(name)s.commit() didn't raise EmptyCommit."
                    % vars(self.Class))
            except EmptyCommit:
                pass

        def test_empty_commit_allowed(self):
            """Empty commit should _not_ raise exception if allow_empty=True.
            """
            self._setup_for_empty_commit()
            self.s.commit(self.commit_msg, self.commit_body,
                          allow_empty=True)

        def test_commit_revision_ids(self):
            """Commit / revision_id should agree on revision ids.
            """
            def val(i):
                return '%s:%d' % (self.val, i+1)
            self.s.add(self.id, directory=False)
            revs = []
            for i in range(10):
                self.s.set(self.id, val(i))
                revs.append(self.s.commit('%s: %d' % (self.commit_msg, i),
                                          self.commit_body))
            for i in range(10):
                rev = self.s.revision_id(i+1)
                self.failUnless(rev == revs[i],
                                "%s.revision_id(%d) returned %s not %s"
                                % (vars(self.Class)['name'], i+1, rev, revs[i]))
            for i in range(-1, -9, -1):
                rev = self.s.revision_id(i)
                self.failUnless(rev == revs[i],
                                "%s.revision_id(%d) returned %s not %s"
                                % (vars(self.Class)['name'], i, rev, revs[i]))

        def test_get_previous_version(self):
            """Get should be able to return the previous version.
            """
            def val(i):
                return '%s:%d' % (self.val, i+1)
            self.s.add(self.id, directory=False)
            revs = []
            for i in range(10):
                self.s.set(self.id, val(i))
                revs.append(self.s.commit('%s: %d' % (self.commit_msg, i),
                                          self.commit_body))
            for i in range(10):
                ret = self.s.get(self.id, revision=revs[i])
                self.failUnless(ret == val(i),
                                "%s.get() returned %s not %s for revision %s"
                                % (vars(self.Class)['name'], ret, val(i), revs[i]))

        def test_get_previous_children(self):
            """Children list should be revision dependent.
            """
            self.s.add('parent', directory=True)
            revs = []
            cur_children = []
            children = []
            for i in range(10):
                new_child = str(i)
                self.s.add(new_child, 'parent')
                self.s.set(new_child, self.val)
                revs.append(self.s.commit('%s: %d' % (self.commit_msg, i),
                                          self.commit_body))
                cur_children.append(new_child)
                children.append(list(cur_children))
            for i in range(10):
                ret = sorted(self.s.children('parent', revision=revs[i]))
                self.failUnless(ret == children[i],
                                "%s.children() returned %s not %s for revision %s"
                                % (vars(self.Class)['name'], ret,
                                   children[i], revs[i]))

        def test_avoid_previous_grandchildren(self):
            """Previous grandchildren should not be returned as children.
            """
            self.s.add('parent', directory=True)
            revs = []
            cur_children = []
            children = []
            for i in range(5):
                new_child = 'parent/%s' % str(i)
                directory = (i % 2 == 0)
                self.s.add(new_child, 'parent', directory=directory)
                cur_children.append(new_child)
                children.append(list(cur_children))
                if directory:
                    for j in range(3):
                        new_grandchild = '%s/%s' % (new_child, str(j))
                        directory = (j % 2 == 0)
                        self.s.add(
                            new_grandchild, new_child, directory=directory)
                        if not directory:
                            self.s.set(new_grandchild, self.val)
                else:
                    self.s.set(new_child, self.val)
                revs.append(self.s.commit('%s: %d' % (self.commit_msg, i),
                                          self.commit_body))
            for rev,cur_children in zip(revs, children):
                ret = sorted(self.s.children('parent', revision=rev))
                self.failUnless(ret == cur_children,
                                "%s.children() returned %s not %s for revision %s"
                                % (vars(self.Class)['name'], ret,
                                   cur_children, rev))

    class VersionedStorage_changed_TestCase (VersionedStorageTestCase):
        """Test cases for VersionedStorage.changed() method."""

        def test_changed(self):
            """Changed lists should reflect past activity"""
            self.s.add('dir', directory=True)
            self.s.add('modified', parent='dir')
            self.s.set('modified', 'some value to be modified')
            self.s.add('moved', parent='dir')
            self.s.set('moved', 'this entry will be moved')
            self.s.add('removed', parent='dir')
            self.s.set('removed', 'this entry will be deleted')
            revA = self.s.commit('Initial state')
            self.s.add('new', parent='dir')
            self.s.set('new', 'this entry is new')
            self.s.set('modified', 'a new value')
            self.s.remove('moved')
            self.s.add('moved2', parent='dir')
            self.s.set('moved2', 'this entry will be moved')
            self.s.remove('removed')
            revB = self.s.commit('Final state')
            new,mod,rem = self.s.changed(revA)
            self.failUnless(sorted(new) == ['moved2', 'new'],
                            'Unexpected new: %s' % new)
            self.failUnless(mod == ['modified'],
                            'Unexpected modified: %s' % mod)
            self.failUnless(sorted(rem) == ['moved', 'removed'],
                            'Unexpected removed: %s' % rem)

    def make_storage_testcase_subclasses(storage_class, namespace):
        """Make StorageTestCase subclasses for storage_class in namespace."""
        storage_testcase_classes = [
            c for c in (
                ob for ob in globals().values() if isinstance(ob, type))
            if issubclass(c, StorageTestCase) \
                and c.Class == Storage]

        for base_class in storage_testcase_classes:
            testcase_class_name = storage_class.__name__ + base_class.__name__
            testcase_class_bases = (base_class,)
            testcase_class_dict = dict(base_class.__dict__)
            testcase_class_dict['Class'] = storage_class
            testcase_class = type(
                testcase_class_name, testcase_class_bases, testcase_class_dict)
            setattr(namespace, testcase_class_name, testcase_class)

    def make_versioned_storage_testcase_subclasses(storage_class, namespace):
        """Make VersionedStorageTestCase subclasses for storage_class in namespace."""
        storage_testcase_classes = [
            c for c in (
                ob for ob in globals().values() if isinstance(ob, type))
            if ((issubclass(c, StorageTestCase) \
                     and c.Class == Storage)
                or
                (issubclass(c, VersionedStorageTestCase) \
                     and c.Class == VersionedStorage))]

        for base_class in storage_testcase_classes:
            testcase_class_name = storage_class.__name__ + base_class.__name__
            testcase_class_bases = (base_class,)
            testcase_class_dict = dict(base_class.__dict__)
            testcase_class_dict['Class'] = storage_class
            testcase_class = type(
                testcase_class_name, testcase_class_bases, testcase_class_dict)
            setattr(namespace, testcase_class_name, testcase_class)

    make_storage_testcase_subclasses(VersionedStorage, sys.modules[__name__])

    unitsuite =unittest.TestLoader().loadTestsFromModule(sys.modules[__name__])
    suite = unittest.TestSuite([unitsuite, doctest.DocTestSuite()])



# Copyright (C) 2005-2012 Aaron Bentley <abentley@panoramicfeedback.com>
#                         Chris Ball <cjb@laptop.org>
#                         Gianluca Montecchi <gian@grys.it>
#                         Marien Zwart <marien.zwart@gmail.com>
#                         Thomas Gerigk <tgerigk@gmx.de>
#                         Tim Guirgies <lt.infiltrator@gmail.com>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

import libbe
import libbe.bug
import libbe.command
import libbe.command.util


class Status (libbe.command.Command):
    """Change a bug's status level

    >>> import sys
    >>> import libbe.bugdir
    >>> bd = libbe.bugdir.SimpleBugDir(memory=False)
    >>> io = libbe.command.StringInputOutput()
    >>> io.stdout = sys.stdout
    >>> ui = libbe.command.UserInterface(io=io)
    >>> ui.storage_callbacks.set_storage(bd.storage)
    >>> cmd = Status(ui=ui)
    >>> cmd._storage = bd.storage

    >>> bd.bug_from_uuid('a').status
    'open'
    >>> ret = ui.run(cmd, args=['closed', '/a'])
    >>> bd.flush_reload()
    >>> bd.bug_from_uuid('a').status
    'closed'
    >>> ret = ui.run(cmd, args=['none', '/a'])
    Traceback (most recent call last):
    UserError: Invalid status level: none
    >>> ui.cleanup()
    >>> bd.cleanup()
    """
    name = 'status'

    def __init__(self, *args, **kwargs):
        libbe.command.Command.__init__(self, *args, **kwargs)
        self.args.extend([
                libbe.command.Argument(
                    name='status', metavar='STATUS', default=None,
                    completion_callback=libbe.command.util.complete_status),
                libbe.command.Argument(
                    name='bug-id', metavar='BUG-ID', default=None,
                    repeatable=True,
                    completion_callback=libbe.command.util.complete_bug_id),
                ])

    def _run(self, **params):
        bugdirs = self._get_bugdirs()
        for bug_id in params['bug-id']:
            bugdir,bug,comment = (
                libbe.command.util.bugdir_bug_comment_from_user_id(
                    bugdirs, bug_id))
            if bug.status != params['status']:
                try:
                    bug.status = params['status']
                except ValueError, e:
                    if e.name != 'status':
                        raise e
                    raise libbe.command.UserError(
                        'Invalid status level: %s' % e.value)
        return 0

    def _long_help(self):
        try: # See if there are any per-tree status configurations
            bugdirs = self._get_bugdirs()
        except NotImplementedError:
            pass # No tree, just show the defaults
        longest_status_len = max([len(s) for s in libbe.bug.status_values])
        active_statuses = []
        for status in libbe.bug.active_status_values :
            description = libbe.bug.status_description[status]
            s = '%*s : %s' % (longest_status_len, status, description)
            active_statuses.append(s)
        inactive_statuses = []
        for status in libbe.bug.inactive_status_values :
            description = libbe.bug.status_description[status]
            s = '%*s : %s' % (longest_status_len, status, description)
            inactive_statuses.append(s)
        ret = """
Show or change a bug's status.

If no status is specified, the current value is printed.  If a status
is specified, it will be assigned to the bug.

There are two classes of statuses, active and inactive, which are only
important for commands like "be list" that show only active bugs by
default.

Active status levels are:
  %s
Inactive status levels are:
  %s

You can overide the list of allowed statuses on a per-repository
basis.  See `be set --help` for details.
""" % ('\n  '.join(active_statuses), '\n  '.join(inactive_statuses))
        return ret



# Copyright (C) 2005-2012 Aaron Bentley <abentley@panoramicfeedback.com>
#                         Chris Ball <cjb@laptop.org>
#                         Gianluca Montecchi <gian@grys.it>
#                         Marien Zwart <marien.zwart@gmail.com>
#                         Thomas Gerigk <tgerigk@gmx.de>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

import libbe
import libbe.command
import libbe.command.util


TOPICS = {
    'repo': """A BE repository containing child bugdirs

BE repositories are stored in an abstract `Storage` instance, which
may or may not be versioned.  If you're using BE to track bugs in your
local software, you'll probably be using an on-disk storage based on
the VCS you use to version the storage.  See `be help init` for
details about automatic VCS-detection.

While most users will be using local storage, BE also supports remote
storage servers.  This allows projects to publish their local
repository in a way that's directly accessible to remote users.  The
remote users can then use a local BE client to interact with the
remote repository, without having to create a local copy of the
repository.  The remote server will be running something like:

    $ be serve-storage --host 123.123.123.123 --port 54321

And the local client can run:

    $ be --repo http://123.123.123.123:54321 list

or whichever command they like.        

Because the storage server serves repositories at the `Storage` level,
it can be inefficient.  For example, `be list` will have to transfer
the data for all the bugs in a repository over the wire.  The storage
server can also be harder to lock down, because users with write
access can potentially store data that cannot be parsed by BE.  For a
more efficient server, see `be serve-commands`.
""",
##
    'server': """A server for remote BE command execution

The usual way for a user to interact with a BE bug tracker for a
particular project is to clone the project repository.  They can then
use their local BE client to browse the repository and make changes,
before pushing their changes back upstream.  For the average user
seeking to file a bug or comment, this can be too much work.  One way
to simplify the process is to use a storage server (see `be help
repo`), but this is not always ideal.  A more robust approach is to
use a command server.

The remote server will be running something like:

    $ be serve-commands --host 123.123.123.123 --port 54321

And the local client can run:

    $ be --server http://123.123.123.123:54321 list

or whichever command they like.  The command line arguments are parsed
locally, and then POSTed to the command server, where the command is
executed.  The output of the command is returned to the client for
display.  This requires much less traffic over the wire than running
the same command via a storage server.
""",
        }


class Help (libbe.command.Command):
    """Print help for given command or topic

    >>> import sys
    >>> import libbe.bugdir
    >>> io = libbe.command.StringInputOutput()
    >>> io.stdout = sys.stdout
    >>> ui = libbe.command.UserInterface(io=io)
    >>> cmd = Help()

    >>> ret = ui.run(cmd, args=['help'])
    usage: be help [options] [TOPIC]
    <BLANKLINE>
    Options:
      -h, --help  Print a help message.
    <BLANKLINE>
      --complete  Print a list of possible completions.
    <BLANKLINE>
    <BLANKLINE>
    Print help for specified command/topic or list of all commands.
    """
    name = 'help'

    def __init__(self, *args, **kwargs):
        libbe.command.Command.__init__(self, *args, **kwargs)
        self.args.extend([
                libbe.command.Argument(
                    name='topic', metavar='TOPIC', default=None,
                    optional=True,
                    completion_callback=self.complete_topic)
                ])

    def _run(self, **params):
        if params['topic'] == None:
            if hasattr(self.ui, 'help'):
                print >> self.stdout, self.ui.help().rstrip('\n')
        elif params['topic'] in libbe.command.commands(command_names=True):
            module = libbe.command.get_command(params['topic'])
            Class = libbe.command.get_command_class(module,params['topic'])
            c = Class(ui=self.ui)
            self.ui.setup_command(c)
            print >> self.stdout, c.help().rstrip('\n')
        elif params['topic'] in TOPICS:
            print >> self.stdout, TOPICS[params['topic']].rstrip('\n')
        else:
            raise libbe.command.UserError(
                '"%s" is neither a command nor topic' % params['topic'])
        return 0

    def _long_help(self):
        return """
Print help for specified command/topic or list of all commands.
"""

    def complete_topic(self, command, argument, fragment=None):
        commands = libbe.command.util.complete_command()
        topics = sorted(TOPICS.keys())
        return commands + topics



# Copyright (C) 2009-2012 Chris Ball <cjb@laptop.org>
#                         Gianluca Montecchi <gian@grys.it>
#                         Robert Lehmann <mail@robertlehmann.de>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

import copy
import itertools
import os

import libbe
import libbe.bug
import libbe.bugdir
import libbe.command
import libbe.command.util
import libbe.util.tree

BLOCKS_TAG="BLOCKS:"
BLOCKED_BY_TAG="BLOCKED-BY:"


class Filter (object):
    def __init__(self, status='all', severity='all', assigned='all',
                 target='all', extra_strings_regexps=[]):
        self.status = status
        self.severity = severity
        self.assigned = assigned
        self.target = target
        self.extra_strings_regexps = extra_strings_regexps

    def __call__(self, bugdirs, bug):
        if self.status != 'all' and not bug.status in self.status:
            return False
        if self.severity != 'all' and not bug.severity in self.severity:
            return False
        if self.assigned != 'all' and not bug.assigned in self.assigned:
            return False
        if self.target == 'all':
            pass
        else:
            target_bug = libbe.command.target.bug_target(bugdirs, bug)
            if self.target in ['none', None]:
                if target_bug.summary != None:
                    return False
            else:
                if target_bug.summary != self.target:
                    return False
        if len(bug.extra_strings) == 0:
            if len(self.extra_strings_regexps) > 0:
                return False
        elif len(self.extra_strings_regexps) > 0:
            matched = False
            for string in bug.extra_strings:
                for regexp in self.extra_strings_regexps:
                    if regexp.match(string):
                        matched = True
                        break
                if matched == True:
                    break
            if matched == False:
                return False
        return True

def parse_status(status):
    if status == 'all':
        status = libbe.bug.status_values
    elif status == 'active':
        status = list(libbe.bug.active_status_values)
    elif status == 'inactive':
        status = list(libbe.bug.inactive_status_values)
    else:
        status = libbe.command.util.select_values(
            status, libbe.bug.status_values)
    return status

def parse_severity(severity, important=False):
    if important == True:
        serious = libbe.bug.severity_values.index('serious')
        severity = list(libbe.bug.severity_values[serious:])
    elif severity == 'all':
        severity = libbe.bug.severity_values
    else:
        severity = libbe.command.util.select_values(
            severity, libbe.bug.severity_values)
    return severity


class BrokenLink (Exception):
    def __init__(self, blocked_bug, blocking_bug, blocks=True):
        if blocks == True:
            msg = "Missing link: %s blocks %s" \
                % (blocking_bug.id.user(), blocked_bug.id.user())
        else:
            msg = "Missing link: %s blocked by %s" \
                % (blocked_bug.id.user(), blocking_bug.id.user())
        Exception.__init__(self, msg)
        self.blocked_bug = blocked_bug
        self.blocking_bug = blocking_bug

class Depend (libbe.command.Command):
    """Add/remove bug dependencies

    >>> import sys
    >>> bd = libbe.bugdir.SimpleBugDir(memory=False)
    >>> io = libbe.command.StringInputOutput()
    >>> io.stdout = sys.stdout
    >>> ui = libbe.command.UserInterface(io=io)
    >>> ui.storage_callbacks.set_storage(bd.storage)
    >>> cmd = Depend(ui=ui)

    >>> ret = ui.run(cmd, args=['/a', '/b'])
    abc/a blocked by:
    abc/b
    >>> ret = ui.run(cmd, args=['/a'])
    abc/a blocked by:
    abc/b
    >>> ret = ui.run(cmd, {'show-status':True}, ['/a']) # doctest: +NORMALIZE_WHITESPACE
    abc/a blocked by:
    abc/b closed
    >>> ret = ui.run(cmd, args=['/b', '/a'])
    abc/b blocked by:
    abc/a
    abc/b blocks:
    abc/a
    >>> ret = ui.run(cmd, {'show-status':True}, ['/a']) # doctest: +NORMALIZE_WHITESPACE
    abc/a blocked by:
    abc/b closed
    abc/a blocks:
    abc/b closed
    >>> ret = ui.run(cmd, {'show-summary':True}, ['/a']) # doctest: +NORMALIZE_WHITESPACE
    abc/a blocked by:
    abc/b       Bug B
    abc/a blocks:
    abc/b       Bug B
    >>> ret = ui.run(cmd, {'repair':True})
    >>> ret = ui.run(cmd, {'remove':True}, ['/b', '/a'])
    abc/b blocks:
    abc/a
    >>> ret = ui.run(cmd, {'remove':True}, ['/a', '/b'])
    >>> ui.cleanup()
    >>> bd.cleanup()
    """
    name = 'depend'

    def __init__(self, *args, **kwargs):
        libbe.command.Command.__init__(self, *args, **kwargs)
        self.options.extend([
                libbe.command.Option(name='remove', short_name='r',
                    help='Remove dependency (instead of adding it)'),
                libbe.command.Option(name='show-status', short_name='s',
                    help='Show status of blocking bugs'),
                libbe.command.Option(name='show-summary', short_name='S',
                    help='Show summary of blocking bugs'),
                libbe.command.Option(name='status',
                    help='Only show bugs matching the STATUS specifier',
                    arg=libbe.command.Argument(
                        name='status', metavar='STATUS', default=None,
                        completion_callback=libbe.command.util.complete_status)),
                libbe.command.Option(name='severity',
                    help='Only show bugs matching the SEVERITY specifier',
                    arg=libbe.command.Argument(
                        name='severity', metavar='SEVERITY', default=None,
                        completion_callback=libbe.command.util.complete_severity)),
                libbe.command.Option(name='tree-depth', short_name='t',
                    help='Print dependency tree rooted at BUG-ID with DEPTH levels of both blockers and blockees.  Set DEPTH <= 0 to disable the depth limit.',
                    arg=libbe.command.Argument(
                        name='tree-depth', metavar='INT', type='int',
                        completion_callback=libbe.command.util.complete_severity)),
                libbe.command.Option(name='repair',
                    help='Check for and repair one-way links'),
                ])
        self.args.extend([
                libbe.command.Argument(
                    name='bug-id', metavar='BUG-ID', default=None,
                    optional=True,
                    completion_callback=libbe.command.util.complete_bug_id),
                libbe.command.Argument(
                    name='blocking-bug-id', metavar='BUG-ID', default=None,
                    optional=True,
                    completion_callback=libbe.command.util.complete_bug_id),
                ])

    def _run(self, **params):
        if params['repair'] == True and params['bug-id'] != None:
            raise libbe.command.UserError(
                'No arguments with --repair calls.')
        if params['repair'] == False and params['bug-id'] == None:
            raise libbe.command.UserError(
                'Must specify either --repair or a BUG-ID')
        if params['tree-depth'] != None \
                and params['blocking-bug-id'] != None:
            raise libbe.command.UserError(
                'Only one bug id used in tree mode.')
        bugdirs = self._get_bugdirs()
        if params['repair'] == True:
            good,fixed,broken = check_dependencies(
                bugdirs, repair_broken_links=True)
            assert len(broken) == 0, broken
            if len(fixed) > 0:
                print >> self.stdout, 'Fixed the following links:'
                print >> self.stdout, \
                    '\n'.join(['%s |-- %s' % (blockee.id.user(), blocker.id.user())
                               for blockee,blocker in fixed])
            return 0
        status = parse_status(params['status'])
        severity = parse_severity(params['severity'])
        filter = Filter(status, severity)

        bugdir,bugA,dummy_comment = (
            libbe.command.util.bugdir_bug_comment_from_user_id(
                bugdirs, params['bug-id']))

        if params['tree-depth'] != None:
            dtree = DependencyTree(bugdirs, bugA, params['tree-depth'], filter)
            if len(dtree.blocked_by_tree()) > 0:
                print >> self.stdout, '%s blocked by:' % bugA.id.user()
                for depth,node in dtree.blocked_by_tree().thread():
                    if depth == 0: continue
                    print >> self.stdout, (
                        '%s%s'
                        % (' '*(depth), self.bug_string(node.bug, params)))
            if len(dtree.blocks_tree()) > 0:
                print >> self.stdout, '%s blocks:' % bugA.id.user()
                for depth,node in dtree.blocks_tree().thread():
                    if depth == 0: continue
                    print >> self.stdout, (
                        '%s%s'
                        % (' '*(depth), self.bug_string(node.bug, params)))
            return 0

        if params['blocking-bug-id'] != None:
            bugdirB,bugB,dummy_comment = (
                libbe.command.util.bugdir_bug_comment_from_user_id(
                    bugdirs, params['blocking-bug-id']))
            if params['remove'] == True:
                remove_block(bugA, bugB)
            else: # add the dependency
                add_block(bugA, bugB)

        blocked_by = get_blocked_by(bugdirs, bugA)

        if len(blocked_by) > 0:
            print >> self.stdout, '%s blocked by:' % bugA.id.user()
            print >> self.stdout, \
                '\n'.join([self.bug_string(_bug, params)
                           for _bug in blocked_by])
        blocks = get_blocks(bugdirs, bugA)
        if len(blocks) > 0:
            print >> self.stdout, '%s blocks:' % bugA.id.user()
            print >> self.stdout, \
                '\n'.join([self.bug_string(_bug, params)
                           for _bug in blocks])
        return 0

    def bug_string(self, _bug, params):
        fields = [_bug.id.user()]
        if params['show-status'] == True:
            fields.append(_bug.status)
        if params['show-summary'] == True:
            fields.append(_bug.summary)
        return '\t'.join(fields)

    def _long_help(self):
        return """
Set a dependency with the second bug (B) blocking the first bug (A).
If bug B is not specified, just print a list of bugs blocking (A).

To search for bugs blocked by a particular bug, try
  $ be list --extra-strings BLOCKED-BY:<your-bug-uuid>

The --status and --severity options allow you to either blacklist or
whitelist values, for example
  $ be list --status open,assigned
will only follow and print dependencies with open or assigned status.
You select blacklist mode by starting the list with a minus sign, for
example
  $ be list --severity -target
which will only follow and print dependencies with non-target severity.

If neither bug A nor B is specified, check for and repair the missing
side of any one-way links.

The "|--" symbol in the repair-mode output is inspired by the
"negative feedback" arrow common in biochemistry.  See, for example
  http://www.nature.com/nature/journal/v456/n7223/images/nature07513-f5.0.jpg
"""

# internal helper functions

def _generate_blocks_string(blocked_bug):
    return '%s%s' % (BLOCKS_TAG, blocked_bug.uuid)

def _generate_blocked_by_string(blocking_bug):
    return '%s%s' % (BLOCKED_BY_TAG, blocking_bug.uuid)

def _parse_blocks_string(string):
    assert string.startswith(BLOCKS_TAG)
    return string[len(BLOCKS_TAG):]

def _parse_blocked_by_string(string):
    assert string.startswith(BLOCKED_BY_TAG)
    return string[len(BLOCKED_BY_TAG):]

def _add_remove_extra_string(bug, string, add):
    estrs = bug.extra_strings
    if add == True:
        estrs.append(string)
    else: # remove the string
        estrs.remove(string)
    bug.extra_strings = estrs # reassign to notice change

def _get_blocks(bug):
    uuids = []
    for line in bug.extra_strings:
        if line.startswith(BLOCKS_TAG):
            uuids.append(_parse_blocks_string(line))
    return uuids

def _get_blocked_by(bug):
    uuids = []
    for line in bug.extra_strings:
        if line.startswith(BLOCKED_BY_TAG):
            uuids.append(_parse_blocked_by_string(line))
    return uuids

def _repair_one_way_link(blocked_bug, blocking_bug, blocks=None):
    if blocks == True: # add blocks link
        blocks_string = _generate_blocks_string(blocked_bug)
        _add_remove_extra_string(blocking_bug, blocks_string, add=True)
    else: # add blocked by link
        blocked_by_string = _generate_blocked_by_string(blocking_bug)
        _add_remove_extra_string(blocked_bug, blocked_by_string, add=True)

# functions exposed to other modules

def add_block(blocked_bug, blocking_bug):
    blocked_by_string = _generate_blocked_by_string(blocking_bug)
    _add_remove_extra_string(blocked_bug, blocked_by_string, add=True)
    blocks_string = _generate_blocks_string(blocked_bug)
    _add_remove_extra_string(blocking_bug, blocks_string, add=True)

def remove_block(blocked_bug, blocking_bug):
    blocked_by_string = _generate_blocked_by_string(blocking_bug)
    _add_remove_extra_string(blocked_bug, blocked_by_string, add=False)
    blocks_string = _generate_blocks_string(blocked_bug)
    _add_remove_extra_string(blocking_bug, blocks_string, add=False)

def get_blocks(bugdirs, bug):
    """
    Return a list of bugs that the given bug blocks.
    """
    blocks = []
    for uuid in _get_blocks(bug):
        blocks.append(libbe.command.util.bug_from_uuid(bugdirs, uuid))
    return blocks

def get_blocked_by(bugdirs, bug):
    """
    Return a list of bugs blocking the given bug.
    """
    blocked_by = []
    for uuid in _get_blocked_by(bug):
        blocked_by.append(libbe.command.util.bug_from_uuid(bugdirs, uuid))
    return blocked_by

def check_dependencies(bugdirs, repair_broken_links=False):
    """
    Check that links are bi-directional for all bugs in bugdir.

    >>> import libbe.bugdir
    >>> bugdir = libbe.bugdir.SimpleBugDir()
    >>> bugdirs = {bugdir.uuid: bugdir}
    >>> a = bugdir.bug_from_uuid('a')
    >>> b = bugdir.bug_from_uuid('b')
    >>> blocked_by_string = _generate_blocked_by_string(b)
    >>> _add_remove_extra_string(a, blocked_by_string, add=True)
    >>> good,repaired,broken = check_dependencies(
    ...     bugdirs, repair_broken_links=False)
    >>> good
    []
    >>> repaired
    []
    >>> broken
    [(Bug(uuid='a'), Bug(uuid='b'))]
    >>> _get_blocks(b)
    []
    >>> good,repaired,broken = check_dependencies(
    ...     bugdirs, repair_broken_links=True)
    >>> _get_blocks(b)
    ['a']
    >>> good
    []
    >>> repaired
    [(Bug(uuid='a'), Bug(uuid='b'))]
    >>> broken
    []
    >>> bugdir.cleanup()
    """
    for bugdir in bugdirs.values():
        if bugdir.storage is not None:
            bugdir.load_all_bugs()
    good_links = []
    fixed_links = []
    broken_links = []
    for bugdir in bugdirs.values():
        for bug in bugdir:
            for blocker in get_blocked_by(bugdirs, bug):
                blocks = get_blocks(bugdirs, blocker)
                if (bug, blocks) in good_links+fixed_links+broken_links:
                    continue # already checked that link
                if bug not in blocks:
                    if repair_broken_links == True:
                        _repair_one_way_link(bug, blocker, blocks=True)
                        fixed_links.append((bug, blocker))
                    else:
                        broken_links.append((bug, blocker))
                else:
                    good_links.append((bug, blocker))
            for blockee in get_blocks(bugdirs, bug):
                blocked_by = get_blocked_by(bugdirs, blockee)
                if (blockee, bug) in good_links+fixed_links+broken_links:
                    continue # already checked that link
                if bug not in blocked_by:
                    if repair_broken_links == True:
                        _repair_one_way_link(blockee, bug, blocks=False)
                        fixed_links.append((blockee, bug))
                    else:
                        broken_links.append((blockee, bug))
                else:
                    good_links.append((blockee, bug))
    return (good_links, fixed_links, broken_links)

class DependencyTree (object):
    """
    Note: should probably be DependencyDiGraph.
    """
    def __init__(self, bugdirs, root_bug, depth_limit=0, filter=None):
        self.bugdirs = bugdirs
        self.root_bug = root_bug
        self.depth_limit = depth_limit
        self.filter = filter

    def _build_tree(self, child_fn):
        root = libbe.util.tree.Tree()
        root.bug = self.root_bug
        root.depth = 0
        stack = [root]
        while len(stack) > 0:
            node = stack.pop()
            if self.depth_limit > 0 and node.depth == self.depth_limit:
                continue
            for bug in child_fn(self.bugdirs, node.bug):
                if not self.filter(self.bugdirs, bug):
                    continue
                child = libbe.util.tree.Tree()
                child.bug = bug
                child.depth = node.depth+1
                node.append(child)
                stack.append(child)
        return root

    def blocks_tree(self):
        if not hasattr(self, "_blocks_tree"):
            self._blocks_tree = self._build_tree(get_blocks)
        return self._blocks_tree

    def blocked_by_tree(self):
        if not hasattr(self, "_blocked_by_tree"):
            self._blocked_by_tree = self._build_tree(get_blocked_by)
        return self._blocked_by_tree



# Copyright (C) 2005-2012 Aaron Bentley <abentley@panoramicfeedback.com>
#                         Chris Ball <cjb@laptop.org>
#                         Gianluca Montecchi <gian@grys.it>
#                         Marien Zwart <marien.zwart@gmail.com>
#                         Thomas Gerigk <tgerigk@gmx.de>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.


import textwrap

import libbe
import libbe.bugdir
import libbe.command
import libbe.command.util
from libbe.storage.util.settings_object import EMPTY


class Set (libbe.command.Command):
    """Change bug directory settings

    >>> import sys
    >>> import libbe.bugdir
    >>> bd = libbe.bugdir.SimpleBugDir(memory=False)
    >>> io = libbe.command.StringInputOutput()
    >>> io.stdout = sys.stdout
    >>> ui = libbe.command.UserInterface(io=io)
    >>> ui.storage_callbacks.set_storage(bd.storage)
    >>> cmd = Set(ui=ui)

    >>> ret = ui.run(cmd, args=['target'])
    None
    >>> ret = ui.run(cmd, args=['target', 'abcdefg'])
    >>> ret = ui.run(cmd, args=['target'])
    abcdefg
    >>> ret = ui.run(cmd, args=['target', 'none'])
    >>> ret = ui.run(cmd, args=['target'])
    None
    >>> ui.cleanup()
    >>> bd.cleanup()
    """
    name = 'set'

    def __init__(self, *args, **kwargs):
        libbe.command.Command.__init__(self, *args, **kwargs)
        self.options.extend([
                libbe.command.Option(name='bugdir', short_name='b',
                    help='Short bugdir UUID to act on.  You '
                    'only need to set this if you have multiple bugdirs in '
                    'your repository.',
                    arg=libbe.command.Argument(
                        name='bugdir', metavar='ID', default=None,
                        completion_callback=libbe.command.util.complete_bugdir_id)),
                ])
        self.args.extend([
                libbe.command.Argument(
                    name='setting', metavar='SETTING', optional=True,
                    completion_callback=complete_bugdir_settings),
                libbe.command.Argument(
                    name='value', metavar='VALUE', optional=True)
                ])

    def _run(self, **params):
        bugdirs = self._get_bugdirs()
        if params['bugdir']:
            bugdir = bugdirs[bugdir]
        elif len(bugdirs) == 1:
            bugdir = bugdirs.values()[0]
        else:
            raise libbe.command.UserError(
                'Ambiguous bugdir {}'.format(sorted(bugdirs.values())))
        if params['setting'] == None:
            keys = bugdir.settings_properties
            keys.sort()
            for key in keys:
                print >> self.stdout, \
                    '%16s: %s' % (key, _value_string(bugdir, key))
            return 0
        if params['setting'] not in bugdir.settings_properties:
            msg = 'Invalid setting %s\n' % params['setting']
            msg += 'Allowed settings:\n  '
            msg += '\n  '.join(bugdir.settings_properties)
            raise libbe.command.UserError(msg)
        if params['value'] == None:
            print _value_string(bugdir, params['setting'])
        else:
            if params['value'] == 'none':
                params['value'] = EMPTY
            old_setting = bugdir.settings.get(params['setting'])
            attr = bugdir._setting_name_to_attr_name(params['setting'])
            setattr(bugdir, attr, params['value'])
        return 0

    def _long_help(self):
        return """
Show or change per-tree settings.

If name and value are supplied, the name is set to a new value.
If no value is specified, the current value is printed.
If no arguments are provided, all names and values are listed.

To unset a setting, set it to "none".

Allowed settings are:

%s

Note that this command does not provide a good interface for some of
these settings (yet!).  You may need to edit the bugdir settings file
(`.be/<bugdir>/settings`) manually.  Examples for each troublesome
setting are given below.

Add the following lines to override the default severities and use
your own:

  severities:
    - - target
      - The issue is a target or milestone, not a bug.
    - - wishlist
      - A feature that could improve usefulness, but not a bug.

You may add as many name/description pairs as you wish to have; they
are sorted in order from least important at the top, to most important
at the bottom.  The target severity gets special handling by `be
target`.

Note that the values here _override_ the defaults. That means that if
you like the defaults, and wish to keep them, you will have to copy
them here before adding any of your own.  See `be severity --help` for
the current list.

Add the following lines to override the default statuses and use your
own:

  active_status:
    - - unconfirmed
      - A possible bug which lacks independent existance confirmation.
    - - open
      - A working bug that has not been assigned to a developer.

  inactive_status:
    - - closed
      - The bug is no longer relevant.
    - - fixed
      - The bug should no longer occur.

You may add as many name/description pairs as you wish to have; they
are sorted in order from most important at the top, to least important
at the bottom.

Note that the values here _override_ the defaults. That means that if
you like the defaults, and wish to keep them, you will have to copy
them here before adding any of your own.  See `be status --help` for
the current list.
""" % ('\n'.join(get_bugdir_settings()),)

def get_bugdir_settings():
    settings = []
    for s in libbe.bugdir.BugDir.settings_properties:
        settings.append(s)
    settings.sort()
    documented_settings = []
    for s in settings:
        set = getattr(libbe.bugdir.BugDir, s)
        dstr = set.__doc__.strip()
        # per-setting comment adjustments
        if s == 'vcs_name':
            lines = dstr.split('\n')
            while lines[0].startswith('This property defaults to') == False:
                lines.pop(0)
            assert len(lines) != None, \
                'Unexpected vcs_name docstring:\n  "%s"' % dstr
            lines.insert(
                0, 'The name of the revision control system to use.\n')
            dstr = '\n'.join(lines)
        doc = textwrap.wrap(dstr, width=70, initial_indent='  ',
                            subsequent_indent='  ')
        documented_settings.append('%s\n%s' % (s, '\n'.join(doc)))
    return documented_settings

def _value_string(bugdir, setting):
    val = bugdir.settings.get(setting, EMPTY)
    if val == EMPTY:
        default = getattr(bugdir, bugdir._setting_name_to_attr_name(setting))
        if default not in [None, EMPTY]:
            val = 'None (%s)' % default
        else:
            val = None
    return str(val)

def complete_bugdir_settings(command, argument, fragment=None):
    """
    List possible command completions for fragment.

    Neither the command nor argument arguments are used.
    """
    return libbe.bugdir.BugDir.settings_properties



# Copyright (C) 2009-2012 Chris Ball <cjb@laptop.org>
#                         Gianluca Montecchi <gian@grys.it>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

"""Darcs_ backend.

.. _Darcs: http://darcs.net/
"""

import codecs
import os
import re
import shutil
import sys
import time # work around http://mercurial.selenic.com/bts/issue618
import types
try: # import core module, Python >= 2.5
    from xml.etree import ElementTree
except ImportError: # look for non-core module
    from elementtree import ElementTree
from xml.sax.saxutils import unescape

import libbe
from ...util.subproc import CommandError
from . import base

if libbe.TESTING == True:
    import doctest
    import unittest


def new():
    return Darcs()

class Darcs(base.VCS):
    """:py:class:`base.VCS` implementation for Darcs.
    """
    name='darcs'
    client='darcs'

    def __init__(self, *args, **kwargs):
        base.VCS.__init__(self, *args, **kwargs)
        self.versioned = True
        self.__updated = [] # work around http://mercurial.selenic.com/bts/issue618

    def _vcs_version(self):
        try:
            status,output,error = self._u_invoke_client('--version')
        except CommandError:  # command not found?
            return None
        return output.strip()

    def version_cmp(self, *args):
        """Compare the installed Darcs version `V_i` with another version
        `V_o` (given in `*args`).  Returns

           === ===============
            1  if `V_i > V_o`
            0  if `V_i == V_o`
           -1  if `V_i < V_o`
           === ===============

        Examples
        --------

        >>> d = Darcs(repo='.')
        >>> d._version = '2.3.1 (release)'
        >>> d.version_cmp(2,3,1)
        0
        >>> d.version_cmp(2,3,2)
        -1
        >>> d.version_cmp(2,3,0)
        1
        >>> d.version_cmp(3)
        -1
        >>> d._version = '2.0.0pre2'
        >>> d._parsed_version = None
        >>> d.version_cmp(3)
        -1
        >>> d.version_cmp(2,0,1)
        Traceback (most recent call last):
          ...
        NotImplementedError: Cannot parse non-integer portion "0pre2" of Darcs version "2.0.0pre2"
        """
        if not hasattr(self, '_parsed_version') \
                or self._parsed_version == None:
            num_part = self.version().split(' ')[0]
            self._parsed_version = []
            for num in num_part.split('.'):
                try:
                    self._parsed_version.append(int(num))
                except ValueError, e:
                    self._parsed_version.append(num)
        for current,other in zip(self._parsed_version, args):
            if type(current) != types.IntType:
                raise NotImplementedError(
                    'Cannot parse non-integer portion "%s" of Darcs version "%s"'
                    % (current, self.version()))
            c = cmp(current,other)
            if c != 0:
                return c
        return 0

    def _vcs_get_user_id(self):
        # following http://darcs.net/manual/node4.html#SECTION00410030000000000000
        # as of June 22th, 2010
        if self.repo == None:
            return None
        for pref_file in ['author', 'email']:
            for prefs_dir in [os.path.join(self.repo, '_darcs', 'prefs'),
                              os.path.expanduser(os.path.join('~', '.darcs'))]:
                if prefs_dir == None:
                    continue
                pref_path = os.path.join(prefs_dir, pref_file)
                if os.path.exists(pref_path):
                    return self._vcs_get_file_contents(pref_path).strip()
        for env_variable in ['DARCS_EMAIL', 'EMAIL']:
            if env_variable in os.environ:
                return os.environ[env_variable]
        return None

    def _vcs_detect(self, path):
        if self._u_search_parent_directories(path, "_darcs") != None :
            return True
        return False

    def _vcs_root(self, path):
        """Find the root of the deepest repository containing path."""
        # Assume that nothing funny is going on; in particular, that we aren't
        # dealing with a bare repo.
        if os.path.isdir(path) != True:
            path = os.path.dirname(path)
        darcs_dir = self._u_search_parent_directories(path, '_darcs')
        if darcs_dir == None:
            return None
        return os.path.dirname(darcs_dir)

    def _vcs_init(self, path):
        self._u_invoke_client('init', cwd=path)

    def _vcs_destroy(self):
        vcs_dir = os.path.join(self.repo, '_darcs')
        if os.path.exists(vcs_dir):
            shutil.rmtree(vcs_dir)

    def _vcs_add(self, path):
        if os.path.isdir(path):
            return
        if self.version_cmp(0, 9, 10) == 1:
            self._u_invoke_client('add', '--boring', path)
        else:  # really old versions <= 0.9.10 lack --boring
            self._u_invoke_client('add', path)

    def _vcs_remove(self, path):
        if not os.path.isdir(self._u_abspath(path)):
            os.remove(os.path.join(self.repo, path)) # darcs notices removal

    def _vcs_update(self, path):
        self.__updated.append(path) # work around http://mercurial.selenic.com/bts/issue618
        pass # darcs notices changes

    def _vcs_get_file_contents(self, path, revision=None):
        if revision == None:
            return base.VCS._vcs_get_file_contents(self, path, revision)
        if self.version_cmp(2, 0, 0) == 1:
            status,output,error = self._u_invoke_client( \
                'show', 'contents', '--patch', revision, path)
            return output
        # Darcs versions < 2.0.0pre2 lack the 'show contents' command

        patch = self._diff(revision, path=path, unicode_output=False)

        # '--output -' to be supported in GNU patch > 2.5.9
        # but that hasn't been released as of June 30th, 2009.

        # Rewrite path to status before the patch we want
        args=['patch', '--reverse', path]
        status,output,error = self._u_invoke(args, stdin=patch)

        if os.path.exists(os.path.join(self.repo, path)) == True:
            contents = base.VCS._vcs_get_file_contents(self, path)
        else:
            contents = ''

        # Now restore path to it's current incarnation
        args=['patch', path]
        status,output,error = self._u_invoke(args, stdin=patch)
        return contents

    def _vcs_path(self, id, revision):
        return self._u_find_id(id, revision)

    def _vcs_isdir(self, path, revision):
        if self.version_cmp(2, 3, 1) == 1:
            # Sun Nov 15 20:32:06 EST 2009  thomashartman1@gmail.com
            #   * add versioned show files functionality (darcs show files -p 'some patch')
            status,output,error = self._u_invoke_client( \
                'show', 'files', '--no-files', '--patch', revision)
            children = output.rstrip('\n').splitlines()
            rpath = '.'
            children = [self._u_rel_path(c, rpath) for c in children]
            if path in children:
                return True
            return False
        raise NotImplementedError(
            'Darcs versions <= 2.3.1 lack the --patch option for "show files"')

    def _vcs_listdir(self, path, revision):
        if self.version_cmp(2, 3, 1) == 1:
            # Sun Nov 15 20:32:06 EST 2009  thomashartman1@gmail.com
            #   * add versioned show files functionality (darcs show files -p 'some patch')
            # Wed Dec  9 05:42:21 EST 2009  Luca Molteni <volothamp@gmail.com>
            #   * resolve issue835 show file with file directory arguments
            path = path.rstrip(os.path.sep)
            status,output,error = self._u_invoke_client( \
                'show', 'files', '--patch', revision, path)
            files = output.rstrip('\n').splitlines()
            if path == '.':
                descendents = [self._u_rel_path(f, path) for f in files
                               if f != '.']
            else:
                rel_files = [self._u_rel_path(f, path) for f in files]
                descendents = [f for f in rel_files
                               if f != '.' and not f.startswith('..')]
            return [f for f in descendents if f.count(os.path.sep) == 0]
        # Darcs versions <= 2.3.1 lack the --patch option for 'show files'
        raise NotImplementedError

    def _vcs_commit(self, commitfile, allow_empty=False):
        id = self.get_user_id()
        if id == None or '@' not in id:
            id = '%s <%s@invalid.com>' % (id, id)
        args = ['record', '--all', '--author', id, '--logfile', commitfile]
        status,output,error = self._u_invoke_client(*args)
        empty_strings = ['No changes!']
        # work around http://mercurial.selenic.com/bts/issue618
        if self._u_any_in_string(empty_strings, output) == True \
                and len(self.__updated) > 0:
            time.sleep(1)
            for path in self.__updated:
                os.utime(os.path.join(self.repo, path), None)
            status,output,error = self._u_invoke_client(*args)
        self.__updated = []
        # end work around
        if self._u_any_in_string(empty_strings, output) == True:
            if allow_empty == False:
                raise base.EmptyCommit()
            # note that darcs does _not_ make an empty revision.
            # this returns the last non-empty revision id...
            revision = self._vcs_revision_id(-1)
        else:
            revline = re.compile("Finished recording patch '(.*)'")
            match = revline.search(output)
            assert match != None, output+error
            assert len(match.groups()) == 1
            revision = match.groups()[0]
        return revision

    def _revisions(self):
        """
        Return a list of revisions in the repository.
        """
        status,output,error = self._u_invoke_client('changes', '--xml')
        revisions = []
        xml_str = output.encode('unicode_escape').replace(r'\n', '\n')
        element = ElementTree.XML(xml_str)
        assert element.tag == 'changelog', element.tag
        for patch in element.getchildren():
            assert patch.tag == 'patch', patch.tag
            for child in patch.getchildren():
                if child.tag == 'name':
                    text = unescape(unicode(child.text).decode('unicode_escape').strip())
                    revisions.append(text)
        revisions.reverse()
        return revisions

    def _vcs_revision_id(self, index):
        revisions = self._revisions()
        try:
            if index > 0:
                return revisions[index-1]
            elif index < 0:
                return revisions[index]
            else:
                return None
        except IndexError:
            return None

    def _diff(self, revision, path=None, unicode_output=True):
        revisions = self._revisions()
        i = revisions.index(revision)
        args = ['diff', '--unified']
        if i+1 < len(revisions):
            next_rev = revisions[i+1]
            args.extend(['--from-patch', next_rev])
        if path != None:
            args.append(path)
        kwargs = {'unicode_output':unicode_output}
        status,output,error = self._u_invoke_client(
            *args, **kwargs)
        return output

    def _parse_diff(self, diff_text):
        """_parse_diff(diff_text) -> (new,modified,removed)

        `new`, `modified`, and `removed` are lists of files.

        Example diff text::

          Mon Jan 18 15:19:30 EST 2010  None <None@invalid.com>
            * Final state
          diff -rN --unified old-BEtestgQtDuD/.be/dir/bugs/modified new-BEtestgQtDuD/.be/dir/bugs/modified
          --- old-BEtestgQtDuD/.be/dir/bugs/modified      2010-01-18 15:19:30.000000000 -0500
          +++ new-BEtestgQtDuD/.be/dir/bugs/modified      2010-01-18 15:19:30.000000000 -0500
          @@ -1 +1 @@
          -some value to be modified
          \ No newline at end of file
          +a new value
          \ No newline at end of file
          diff -rN --unified old-BEtestgQtDuD/.be/dir/bugs/moved new-BEtestgQtDuD/.be/dir/bugs/moved
          --- old-BEtestgQtDuD/.be/dir/bugs/moved 2010-01-18 15:19:30.000000000 -0500
          +++ new-BEtestgQtDuD/.be/dir/bugs/moved 1969-12-31 19:00:00.000000000 -0500
          @@ -1 +0,0 @@
          -this entry will be moved
          \ No newline at end of file
          diff -rN --unified old-BEtestgQtDuD/.be/dir/bugs/moved2 new-BEtestgQtDuD/.be/dir/bugs/moved2
          --- old-BEtestgQtDuD/.be/dir/bugs/moved2        1969-12-31 19:00:00.000000000 -0500
          +++ new-BEtestgQtDuD/.be/dir/bugs/moved2        2010-01-18 15:19:30.000000000 -0500
          @@ -0,0 +1 @@
          +this entry will be moved
          \ No newline at end of file
          diff -rN --unified old-BEtestgQtDuD/.be/dir/bugs/new new-BEtestgQtDuD/.be/dir/bugs/new
          --- old-BEtestgQtDuD/.be/dir/bugs/new   1969-12-31 19:00:00.000000000 -0500
          +++ new-BEtestgQtDuD/.be/dir/bugs/new   2010-01-18 15:19:30.000000000 -0500
          @@ -0,0 +1 @@
          +this entry is new
          \ No newline at end of file
          diff -rN --unified old-BEtestgQtDuD/.be/dir/bugs/removed new-BEtestgQtDuD/.be/dir/bugs/removed
          --- old-BEtestgQtDuD/.be/dir/bugs/removed       2010-01-18 15:19:30.000000000 -0500
          +++ new-BEtestgQtDuD/.be/dir/bugs/removed       1969-12-31 19:00:00.000000000 -0500
          @@ -1 +0,0 @@
          -this entry will be deleted
          \ No newline at end of file
          
        """
        new = []
        modified = []
        removed = []
        lines = diff_text.splitlines()
        repodir = os.path.basename(self.repo) + os.path.sep
        i = 0
        while i < len(lines):
            line = lines[i]; i += 1
            if not line.startswith('diff '):
                continue
            file_a,file_b = line.split()[-2:]
            assert file_a.startswith('old-'), \
                'missformed file_a %s' % file_a
            assert file_b.startswith('new-'), \
                'missformed file_a %s' % file_b
            file = file_a[4:]
            assert file_b[4:] == file, \
                'diff file missmatch %s != %s' % (file_a, file_b)
            assert file.startswith(repodir), \
                'missformed file_a %s' % file_a
            file = file[len(repodir):]
            lines_added = 0
            lines_removed = 0
            line = lines[i]; i += 1
            assert line.startswith('--- old-'), \
                'missformed "---" line %s' % line
            time_a = line.split('\t')[1]
            line = lines[i]; i += 1
            assert line.startswith('+++ new-'), \
                'missformed "+++" line %s' % line
            time_b = line.split('\t')[1]
            zero_time = time.strftime('%Y-%m-%d %H:%M:%S.000000000 ',
                                      time.localtime(0))
            # note that zero_time is missing the trailing timezone offset
            if time_a.startswith(zero_time):
                new.append(file)
            elif time_b.startswith(zero_time):
                removed.append(file)
            else:
                modified.append(file)
        return (new,modified,removed)

    def _vcs_changed(self, revision):
        return self._parse_diff(self._diff(revision))


if libbe.TESTING == True:
    base.make_vcs_testcase_subclasses(Darcs, sys.modules[__name__])

    unitsuite =unittest.TestLoader().loadTestsFromModule(sys.modules[__name__])
    suite = unittest.TestSuite([unitsuite, doctest.DocTestSuite()])



# Copyright (C) 2005-2012 Aaron Bentley <abentley@panoramicfeedback.com>
#                         Chris Ball <cjb@laptop.org>
#                         Gianluca Montecchi <gian@grys.it>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

import os.path

import libbe
import libbe.bugdir
import libbe.command
import libbe.storage

class Init (libbe.command.Command):
    """Create an on-disk bug repository

    >>> import os, sys
    >>> import libbe.storage.vcs
    >>> import libbe.storage.vcs.base
    >>> import libbe.util.utility
    >>> io = libbe.command.StringInputOutput()
    >>> io.stdout = sys.stdout
    >>> ui = libbe.command.UserInterface(io=io)
    >>> cmd = Init()

    >>> dir = libbe.util.utility.Dir()
    >>> vcs = libbe.storage.vcs.vcs_by_name('None')
    >>> vcs.repo = dir.path
    >>> try:
    ...     vcs.connect()
    ... except libbe.storage.ConnectionError:
    ...     'got error'
    'got error'
    >>> ui.storage_callbacks.set_unconnected_storage(vcs)
    >>> ui.run(cmd)
    No revision control detected.
    BE repository initialized.
    >>> bd = libbe.bugdir.BugDir(vcs)
    >>> vcs.disconnect()
    >>> vcs.connect()
    >>> bugdir = libbe.bugdir.BugDir(vcs, from_storage=True)
    >>> vcs.disconnect()
    >>> vcs.destroy()
    >>> dir.cleanup()

    >>> dir = libbe.util.utility.Dir()
    >>> vcs = libbe.storage.vcs.installed_vcs()
    >>> vcs.repo = dir.path
    >>> vcs._vcs_init(vcs.repo)
    >>> ui.storage_callbacks.set_unconnected_storage(vcs)
    >>> if vcs.name in libbe.storage.vcs.base.VCS_ORDER:
    ...     ui.run(cmd) # doctest: +ELLIPSIS
    ... else:
    ...     vcs.init()
    ...     vcs.connect()
    ...     print 'Using ... for revision control.\\nDirectory initialized.'
    Using ... for revision control.
    BE repository initialized.
    >>> vcs.disconnect()
    >>> vcs.connect()
    >>> bugdir = libbe.bugdir.BugDir(vcs, from_storage=True)
    >>> vcs.disconnect()
    >>> vcs.destroy()
    >>> dir.cleanup()
    """
    name = 'init'

    def __init__(self, *args, **kwargs):
        libbe.command.Command.__init__(self, *args, **kwargs)

    def _run(self, **params):
        storage = self._get_unconnected_storage()
        if not os.path.isdir(storage.repo):
            raise libbe.command.UserError(
                'No such directory: %s' % storage.repo)
        try:
            storage.connect()
            raise libbe.command.UserError(
                'Directory already initialized: %s' % storage.repo)
        except libbe.storage.ConnectionError:
            pass
        storage.init()
        storage.connect()
        self.ui.storage_callbacks.set_storage(storage)
        bd = libbe.bugdir.BugDir(storage, from_storage=False)
        self.ui.storage_callbacks.set_bugdirs({bd.uuid: bd})
        if bd.storage.name is not 'None':
            print >> self.stdout, \
                'Using %s for revision control.' % storage.name
        else:
            print >> self.stdout, 'No revision control detected.'
        print >> self.stdout, 'BE repository initialized.'

    def _long_help(self):
        return """
This command initializes Bugs Everywhere support for the specified directory
and all its subdirectories.  It will auto-detect any supported revision control
system.  You can use "be set vcs_name" to change the vcs being used.

The directory defaults to your current working directory, but you can
change that by passing the --repo option to be
  $ be --repo path/to/new/bug/root init

When initialized in a version-controlled directory, BE sinks to the
version-control root.  In that case, the BE repository will be created
under that directory, rather than the current directory or the one
passed in --repo.  Consider the following tree, versioned in Git.
  ~
  `--projectX
     |-- .git
     `-- src
Calling
  ~$ be --repo ./projectX/src init
will create the BE repository rooted in projectX:
  ~
  `--projectX
     |-- .be
     |-- .git
     `-- src
"""



#!/usr/bin/env python

import setuptools # so we can `./setup.py develop`
from distutils.core import setup
import os.path

from libbe import version


_this_dir = os.path.dirname(__file__)

rev_id = version.version_info['revision']
rev_date = version.version_info['date']

data_files = []

man_path = os.path.join('doc', 'man', 'be.1')
if os.path.exists(man_path):
    data_files.append(('share/man/man1', [man_path]))

setup(
    name='bugs-everywhere',
    version='{0}'.format(version.version()),
    maintainer='Thomas Levine',
    maintainer_email='wildebeest@thomaslevine.com',
    url='http://bugseverywhere.org/',
    download_url=(
        'http://downloads.bugseverywhere.org/releases/be-{0}.tar.gz'.format(
            version.version())),
    license='GNU General Public License (GPL)',
    platforms=['all'],
    description='Bugtracker supporting distributed revision control',
    long_description=open(os.path.join(_this_dir, 'README'), 'r').read(),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Framework :: CherryPy',
        'Intended Audience :: Customer Service',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Bug Tracking',
        ],

    packages=['libbe',
              'libbe.command',
              'libbe.storage',
              'libbe.storage.util',
              'libbe.storage.vcs',
              'libbe.ui',
              'libbe.ui.util',
              'libbe.util'],
    scripts=['be'],
    data_files=data_files,
    )



# Copyright (C) 2005-2012 Aaron Bentley <abentley@panoramicfeedback.com>
#                         Chris Ball <cjb@laptop.org>
#                         Gianluca Montecchi <gian@grys.it>
#                         Marien Zwart <marien.zwart@gmail.com>
#                         Thomas Gerigk <tgerigk@gmx.de>
#                         Tim Guirgies <lt.infiltrator@gmail.com>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

import libbe
import libbe.bug
import libbe.command
import libbe.command.util


class Severity (libbe.command.Command):
    """Change a bug's severity level

    >>> import sys
    >>> import libbe.bugdir
    >>> bd = libbe.bugdir.SimpleBugDir(memory=False)
    >>> io = libbe.command.StringInputOutput()
    >>> io.stdout = sys.stdout
    >>> ui = libbe.command.UserInterface(io=io)
    >>> ui.storage_callbacks.set_storage(bd.storage)
    >>> cmd = Severity(ui=ui)

    >>> bd.bug_from_uuid('a').severity
    'minor'
    >>> ret = ui.run(cmd, args=['wishlist', '/a'])
    >>> bd.flush_reload()
    >>> bd.bug_from_uuid('a').severity
    'wishlist'
    >>> ret = ui.run(cmd, args=['none', '/a'])
    Traceback (most recent call last):
    UserError: Invalid severity level: none
    >>> ui.cleanup()
    >>> bd.cleanup()
    """
    name = 'severity'

    def __init__(self, *args, **kwargs):
        libbe.command.Command.__init__(self, *args, **kwargs)
        self.args.extend([
                libbe.command.Argument(
                    name='severity', metavar='SEVERITY', default=None,
                    completion_callback=libbe.command.util.complete_severity),
                libbe.command.Argument(
                    name='bug-id', metavar='BUG-ID', default=None,
                    repeatable=True,
                    completion_callback=libbe.command.util.complete_bug_id),
                ])

    def _run(self, **params):
        bugdirs = self._get_bugdirs()
        for bug_id in params['bug-id']:
            bugdir,bug,comment = (
                libbe.command.util.bugdir_bug_comment_from_user_id(
                    bugdirs, bug_id))
            if bug.severity != params['severity']:
                try:
                    bug.severity = params['severity']
                except ValueError, e:
                    if e.name != 'severity':
                        raise e
                    raise libbe.command.UserError(
                        'Invalid severity level: %s' % e.value)
        return 0

    def _long_help(self):
        try: # See if there are any per-tree severity configurations
            bugdirs = self._get_bugdirs()
        except NotImplementedError:
            pass # No tree, just show the defaults
        longest_severity_len = max([len(s) for s in libbe.bug.severity_values])
        severity_levels = []
        for severity in libbe.bug.severity_values :
            description = libbe.bug.severity_description[severity]
            s = '%*s : %s' % (longest_severity_len, severity, description)
            severity_levels.append(s)
        ret = """
Show or change a bug's severity level.

If no severity is specified, the current value is printed.  If a severity level
is specified, it will be assigned to the bug.

Severity levels are:
  %s

You can overide the list of allowed severities on a per-repository
basis.  See `be set --help` for details.
""" % ('\n  '.join(severity_levels))
        return ret



# Copyright (C) 2009-2012 Chris Ball <cjb@laptop.org>
#                         Phil Schumm <philschumm@gmail.com>
#                         Robert Lehmann <mail@robertlehmann.de>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

import codecs
import optparse
import os.path
import StringIO
import sys
import urlparse

import libbe
import libbe.storage
import libbe.storage.util.mapfile
import libbe.ui.util.user
import libbe.util.encoding
import libbe.util.http
import libbe.util.plugin


class UserError (Exception):
    "An error due to improper BE usage."
    pass


class UsageError (UserError):
    """A serious parsing error due to invalid BE command construction.

    The distinction between `UserError`\s and the more specific
    `UsageError`\s is that when displaying a `UsageError` to the user,
    the user is pointed towards the command usage information.  Use
    the more general `UserError` if you feel that usage information
    would not be particularly enlightening.
    """
    def __init__(self, command=None, command_name=None, message=None):
        super(UsageError, self).__init__(message)
        self.command = command
        if command_name is None and command is not None:
            command_name = command.name
        self.command_name = command_name
        self.message = message


class UnknownCommand (UsageError):
    def __init__(self, command_name, message=None):
        uc_message = "Unknown command '%s'" % command_name
        if message is None:
            message = uc_message
        else:
            message = '%s\n(%s)' % (uc_message, message)
        super(UnknownCommand, self).__init__(
            command_name=command_name,
            message=message)


def get_command(command_name):
    """Retrieves the module for a user command

    >>> try:
    ...     get_command('asdf')
    ... except UnknownCommand, e:
    ...     print e
    Unknown command 'asdf'
    (No module named asdf)
    >>> repr(get_command('list')).startswith("<module 'libbe.command.list' from ")
    True
    """
    try:
        cmd = libbe.util.plugin.import_by_name(
            'libbe.command.%s' % command_name.replace("-", "_"))
    except ImportError, e:
        raise UnknownCommand(command_name, message=unicode(e))
    return cmd

def get_command_class(module=None, command_name=None):
    """Retrieves a command class from a module.

    >>> import_xml_mod = get_command('import-xml')
    >>> import_xml = get_command_class(import_xml_mod, 'import-xml')
    >>> repr(import_xml)
    "<class 'libbe.command.import_xml.Import_XML'>"
    >>> import_xml = get_command_class(command_name='import-xml')
    >>> repr(import_xml)
    "<class 'libbe.command.import_xml.Import_XML'>"
    """
    if module == None:
        module = get_command(command_name)
    try:
        cname = command_name.capitalize().replace('-', '_')
        cmd = getattr(module, cname)
    except ImportError, e:
        raise UnknownCommand(command_name)
    return cmd

def modname_to_command_name(modname):
    """Little hack to replicate
    >>> import sys
    >>> def real_modname_to_command_name(modname):
    ...     mod = libbe.util.plugin.import_by_name(
    ...         'libbe.command.%s' % modname)
    ...     attrs = [getattr(mod, name) for name in dir(mod)]
    ...     commands = []
    ...     for attr_name in dir(mod):
    ...         attr = getattr(mod, attr_name)
    ...         try:
    ...             if issubclass(attr, Command):
    ...                 commands.append(attr)
    ...         except TypeError, e:
    ...             pass
    ...     if len(commands) == 0:
    ...         raise Exception('No Command classes in %s' % dir(mod))
    ...     return commands[0].name
    >>> real_modname_to_command_name('new')
    'new'
    >>> real_modname_to_command_name('import_xml')
    'import-xml'
    """
    return modname.replace('_', '-')

def commands(command_names=False):
    for modname in libbe.util.plugin.modnames('libbe.command'):
        if modname not in ['base', 'util']:
            if command_names == False:
                yield modname
            else:
                yield modname_to_command_name(modname)

class CommandInput (object):
    def __init__(self, name, help=''):
        self.name = name
        self.help = help

    def __str__(self):
        return '<%s %s>' % (self.__class__.__name__, self.name)

    def __repr__(self):
        return self.__str__()

class Argument (CommandInput):
    def __init__(self, metavar=None, default=None, type='string',
                 optional=False, repeatable=False,
                 completion_callback=None, *args, **kwargs):
        CommandInput.__init__(self, *args, **kwargs)
        self.metavar = metavar
        self.default = default
        self.type = type
        self.optional = optional
        self.repeatable = repeatable
        self.completion_callback = completion_callback
        if self.metavar == None:
            self.metavar = self.name.upper()

class Option (CommandInput):
    def __init__(self, callback=None, short_name=None, arg=None,
                 *args, **kwargs):
        CommandInput.__init__(self, *args, **kwargs)
        self.callback = callback
        self.short_name = short_name
        self.arg = arg
        if self.arg == None and self.callback == None:
            # use an implicit boolean argument
            self.arg = Argument(name=self.name, help=self.help,
                                default=False, type='bool')
        self.validate()

    def validate(self):
        if self.arg == None:
            assert self.callback != None, self.name
            return
        assert self.callback == None, '%s: %s' (self.name, self.callback)
        assert self.arg.name == self.name, \
            'Name missmatch: %s != %s' % (self.arg.name, self.name)
        assert self.arg.optional == False, self.name
        assert self.arg.repeatable == False, self.name

    def __str__(self):
        return '--%s' % self.name

    def __repr__(self):
        return '<Option %s>' % self.__str__()

class _DummyParser (optparse.OptionParser):
    def __init__(self, command):
        optparse.OptionParser.__init__(self)
        self.remove_option('-h')
        self.command = command
        self._command_opts = []
        for option in self.command.options:
            self._add_option(option)

    def _add_option(self, option):
        # from libbe.ui.command_line.CmdOptionParser._add_option
        option.validate()
        long_opt = '--%s' % option.name
        if option.short_name != None:
            short_opt = '-%s' % option.short_name
        assert '_' not in option.name, \
            'Non-reconstructable option name %s' % option.name
        kwargs = {'dest':option.name.replace('-', '_'),
                  'help':option.help}
        if option.arg == None or option.arg.type == 'bool':
            kwargs['action'] = 'store_true'
            kwargs['metavar'] = None
            kwargs['default'] = False
        else:
            kwargs['type'] = option.arg.type
            kwargs['action'] = 'store'
            kwargs['metavar'] = option.arg.metavar
            kwargs['default'] = option.arg.default
        if option.short_name != None:
            opt = optparse.Option(short_opt, long_opt, **kwargs)
        else:
            opt = optparse.Option(long_opt, **kwargs)
        #option.takes_value = lambda : option.arg != None
        opt._option = option
        self._command_opts.append(opt)
        self.add_option(opt)

class OptionFormatter (optparse.IndentedHelpFormatter):
    def __init__(self, command):
        optparse.IndentedHelpFormatter.__init__(self)
        self.command = command
    def option_help(self):
        # based on optparse.OptionParser.format_option_help()
        parser = _DummyParser(self.command)
        self.store_option_strings(parser)
        ret = []
        ret.append(self.format_heading('Options'))
        self.indent()
        for option in parser._command_opts:
            ret.append(self.format_option(option))
            ret.append('\n')
        self.dedent()
        # Drop the last '\n', or the header if no options or option groups:
        return ''.join(ret[:-1])

class Command (object):
    """One-line command description here.

    >>> c = Command()
    >>> print c.help()
    usage: be command [options]
    <BLANKLINE>
    Options:
      -h, --help  Print a help message.
    <BLANKLINE>
      --complete  Print a list of possible completions.
    <BLANKLINE>
    A detailed help message.
    """
    user_agent = 'BE-HTTP-Command'

    name = 'command'

    def __init__(self, ui=None, server=None):
        self.ui = ui # calling user-interface
        self.server = server # location of eventual execution
        self.status = None
        self.result = None
        self.restrict_file_access = True
        self.options = [
            Option(name='help', short_name='h',
                help='Print a help message.',
                callback=self.help),
            Option(name='complete',
                help='Print a list of possible completions.',
                callback=self.complete),
                ]
        self.args = []

    def run(self, options=None, args=None):
        self.status = 1 # in case we raise an exception
        params = self._parse_options_args(options, args)
        if params['help'] == True:
            pass
        else:
            params.pop('help')
        if params['complete'] != None:
            pass
        else:
            params.pop('complete')

        if self.server:
            self.status = self._run_remote(**params)
        else:
            self.status = self._run(**params)
        return self.status

    def _parse_options_args(self, options=None, args=None):
        if options == None:
            options = {}
        if args == None:
            args = []
        params = {}
        for option in self.options:
            assert option.name not in params, params[option.name]
            if option.name in options:
                params[option.name] = options.pop(option.name)
            elif option.arg != None:
                params[option.name] = option.arg.default
            else: # non-arg options are flags, set to default flag value
                params[option.name] = False
        assert 'user-id' not in params, params['user-id']
        if 'user-id' in options:
            self._user_id = options.pop('user-id')
        if len(options) > 0:
            raise UserError, 'Invalid option passed to command %s:\n  %s' \
                % (self.name, '\n  '.join(['%s: %s' % (k,v)
                                           for k,v in options.items()]))
        in_optional_args = False
        for i,arg in enumerate(self.args):
            if arg.repeatable == True:
                assert i == len(self.args)-1, arg.name
            if in_optional_args == True:
                assert arg.optional == True, arg.name
            else:
                in_optional_args = arg.optional
            if i < len(args):
                if arg.repeatable == True:
                    params[arg.name] = [args[i]]
                else:
                    params[arg.name] = args[i]
            else:  # no value given
                assert in_optional_args == True, arg.name
                params[arg.name] = arg.default
        if len(args) > len(self.args):  # add some additional repeats
            assert self.args[-1].repeatable == True, self.args[-1].name
            params[self.args[-1].name].extend(args[len(self.args):])
        return params

    def _run(self, **kwargs):
        raise NotImplementedError

    def _run_remote(self, **kwargs):
        data = libbe.storage.util.mapfile.generate({
                'command': self.name,
                'parameters': kwargs,
                }, context=0)
        url = urlparse.urljoin(self.server, 'run')
        page,final_url,info = libbe.util.http.get_post_url(
            url=url, get=False, data=data, agent=self.user_agent)
        self.stdout.write(page)
        return 0

    def help(self, *args):
        return '\n\n'.join([self.usage(),
                            self._option_help(),
                            self._long_help().rstrip('\n')])

    def usage(self):
        usage = 'usage: be %s [options]' % self.name
        num_optional = 0
        for arg in self.args:
            usage += ' '
            if arg.optional == True:
                usage += '['
                num_optional += 1
            usage += arg.metavar
            if arg.repeatable == True:
                usage += ' ...'
        usage += ']'*num_optional
        return usage

    def _option_help(self):
        o = OptionFormatter(self)
        return o.option_help().strip('\n')

    def _long_help(self):
        return "A detailed help message."

    def complete(self, argument=None, fragment=None):
        if argument == None:
            ret = ['--%s' % o.name for o in self.options
                    if o.name != 'complete']
            if len(self.args) > 0 and self.args[0].completion_callback != None:
                ret.extend(self.args[0].completion_callback(self, argument, fragment))
            return ret
        elif argument.completion_callback != None:
            # finish a particular argument
            return argument.completion_callback(self, argument, fragment)
        return [] # the particular argument doesn't supply completion info

    def _check_restricted_access(self, storage, path):
        """
        Check that the file at path is inside bugdir.root.  This is
        important if you allow other users to execute becommands with
        your username (e.g. if you're running be-handle-mail through
        your ~/.procmailrc).  If this check wasn't made, a user could
        e.g.  run
          be commit -b ~/.ssh/id_rsa "Hack to expose ssh key"
        which would expose your ssh key to anyone who could read the
        VCS log.

        >>> class DummyStorage (object): pass
        >>> s = DummyStorage()
        >>> s.repo = os.path.expanduser('~/x/')
        >>> c = Command()
        >>> try:
        ...     c._check_restricted_access(s, os.path.expanduser('~/.ssh/id_rsa'))
        ... except UserError, e:
        ...     assert str(e).startswith('file access restricted!'), str(e)
        ...     print 'we got the expected error'
        we got the expected error
        >>> c._check_restricted_access(s, os.path.expanduser('~/x'))
        >>> c._check_restricted_access(s, os.path.expanduser('~/x/y'))
        >>> c.restrict_file_access = False
        >>> c._check_restricted_access(s, os.path.expanduser('~/.ssh/id_rsa'))
        """
        if self.restrict_file_access == True:
            path = os.path.abspath(path)
            repo = os.path.abspath(storage.repo).rstrip(os.path.sep)
            if path == repo or path.startswith(repo+os.path.sep):
                return
            raise UserError('file access restricted!\n  %s not in %s'
                            % (path, repo))

    def cleanup(self):
        pass

class InputOutput (object):
    def __init__(self, stdin=None, stdout=None):
        self.stdin = stdin
        self.stdout = stdout

    def setup_command(self, command):
        if not hasattr(self.stdin, 'encoding'):
            self.stdin.encoding = libbe.util.encoding.get_input_encoding()
        if not hasattr(self.stdout, 'encoding'):
            self.stdout.encoding = libbe.util.encoding.get_output_encoding()
        command.stdin = self.stdin
        command.stdin.encoding = self.stdin.encoding
        command.stdout = self.stdout
        command.stdout.encoding = self.stdout.encoding

    def cleanup(self):
        pass

class StdInputOutput (InputOutput):
    def __init__(self, input_encoding=None, output_encoding=None):
        stdin,stdout = self._get_io(input_encoding, output_encoding)
        InputOutput.__init__(self, stdin, stdout)

    def _get_io(self, input_encoding=None, output_encoding=None):
        if input_encoding == None:
            input_encoding = libbe.util.encoding.get_input_encoding()
        if output_encoding == None:
            output_encoding = libbe.util.encoding.get_output_encoding()
        stdin = codecs.getreader(input_encoding)(sys.stdin)
        stdin.encoding = input_encoding
        stdout = codecs.getwriter(output_encoding)(sys.stdout)
        stdout.encoding = output_encoding
        return (stdin, stdout)

class StringInputOutput (InputOutput):
    """
    >>> s = StringInputOutput()
    >>> s.set_stdin('hello')
    >>> s.stdin.read()
    'hello'
    >>> s.stdin.read()
    ''
    >>> print >> s.stdout, 'goodbye'
    >>> s.get_stdout()
    'goodbye\\n'
    >>> s.get_stdout()
    ''

    Also works with unicode strings

    >>> s.set_stdin(u'hello')
    >>> s.stdin.read()
    u'hello'
    >>> print >> s.stdout, u'goodbye'
    >>> s.get_stdout()
    u'goodbye\\n'
    """
    def __init__(self):
        stdin = StringIO.StringIO()
        stdin.encoding = 'utf-8'
        stdout = StringIO.StringIO()
        stdout.encoding = 'utf-8'
        InputOutput.__init__(self, stdin, stdout)

    def set_stdin(self, stdin_string):
        self.stdin = StringIO.StringIO(stdin_string)

    def get_stdout(self):
        ret = self.stdout.getvalue()
        self.stdout.truncate(size=0)
        return ret

class UnconnectedStorageGetter (object):
    def __init__(self, location):
        self.location = location

    def __call__(self):
        return libbe.storage.get_storage(self.location)

class StorageCallbacks (object):
    def __init__(self, location=None):
        if location == None:
            location = '.'
        self.location = location
        self._get_unconnected_storage = UnconnectedStorageGetter(location)

    def setup_command(self, command):
        command._get_unconnected_storage = self.get_unconnected_storage
        command._get_storage = self.get_storage
        command._get_bugdirs = self.get_bugdirs

    def get_unconnected_storage(self):
        """
        Callback for use by commands that need it.
        
        The returned Storage instance is may actually be connected,
        but commands that make use of the returned value should only
        make use of non-connected Storage methods.  This is mainly
        intended for the init command, which calls Storage.init().
        """
        if not hasattr(self, '_unconnected_storage'):
            if self._get_unconnected_storage == None:
                raise NotImplementedError
            self._unconnected_storage = self._get_unconnected_storage()
        return self._unconnected_storage

    def set_unconnected_storage(self, unconnected_storage):
        self._unconnected_storage = unconnected_storage

    def get_storage(self):
        """Callback for use by commands that need it."""
        if not hasattr(self, '_storage'):
            self._storage = self.get_unconnected_storage()
            self._storage.connect()
            version = self._storage.storage_version()
            if version != libbe.storage.STORAGE_VERSION:
                raise libbe.storage.InvalidStorageVersion(version)
        return self._storage

    def set_storage(self, storage):
        self._storage = storage

    def get_bugdirs(self):
        """Callback for use by commands that need it."""
        if not hasattr(self, '_bugdirs'):
            storage = self.get_storage()
            self._bugdirs = dict(
                (uuid, libbe.bugdir.BugDir(
                        storage=storage,
                        uuid=uuid,
                        from_storage=True))
                for uuid in storage.children())
        return self._bugdirs

    def set_bugdirs(self, bugdirs):
        self._bugdirs = bugdirs

    def cleanup(self):
        if hasattr(self, '_storage'):
            self._storage.disconnect()

class UserInterface (object):
    def __init__(self, io=None, location=None):
        if io == None:
            io = StringInputOutput()
        self.io = io
        self.storage_callbacks = StorageCallbacks(location)
        self.restrict_file_access = True

    def help(self):
        raise NotImplementedError

    def run(self, command, options=None, args=None):
        self.setup_command(command)
        return command.run(options, args)

    def setup_command(self, command):
        if command.ui is None:
            command.ui = self
        if self.io is not None:
            self.io.setup_command(command)
        if self.storage_callbacks is not None:
            self.storage_callbacks.setup_command(command)        
        command.restrict_file_access = self.restrict_file_access
        command._get_user_id = self._get_user_id

    def _get_user_id(self):
        """Callback for use by commands that need it."""
        if not hasattr(self, '_user_id'):
            self._user_id = libbe.ui.util.user.get_user_id(
                self.storage_callbacks.get_storage())
        return self._user_id

    def cleanup(self):
        if self.storage_callbacks is not None:
            self.storage_callbacks.cleanup()
        self.io.cleanup()



# Copyright (C) 2005-2012 Aaron Bentley <abentley@panoramicfeedback.com>
#                         Chris Ball <cjb@laptop.org>
#                         Gianluca Montecchi <gian@grys.it>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

"""Create, save, and load the per-user config file at :py:func:`path`.
"""

import ConfigParser
import codecs
import os
import os.path

import libbe
import libbe.util.encoding
if libbe.TESTING == True:
    import doctest


default_encoding = libbe.util.encoding.get_text_file_encoding()
"""Default filesystem encoding.

Initialized with :py:func:`libbe.util.encoding.get_text_file_encoding`.
"""

def path():
    """Return the path to the per-user config file.

    Defaults to :file:`~/.config/bugs-everywhere`, but you can
    override the directory with ``XDG_CONFIG_HOME`` from the `XDG Base
    Directory Specification`_.  You can also override the entire path
    by setting the ``BE_CONFIG_PATH`` environment variable.

    .. _XDG Base Directory Specification:
      http://standards.freedesktop.org/basedir-spec/basedir-spec-latest.html
    """
    default_dir = os.path.join('~', '.config')
    dirname = os.path.expanduser(
        os.environ.get('XDG_CONFIG_HOME', default_dir))
    default = os.path.join(dirname, 'bugs-everywhere')
    return os.path.expanduser(os.environ.get('BE_CONFIG_PATH', default))

def set_val(name, value, section="DEFAULT", encoding=None):
    """Set a value in the per-user config file.

    Parameters
    ----------
    name : str
      The name of the value to set.
    value : str or None
      The new value to set (or None to delete the value).
    section : str
      The section to store the name/value in.
    encoding : str
      The config file's encoding, defaults to :py:data:`default_encoding`.
    """
    if encoding == None:
        encoding = default_encoding
    config = ConfigParser.ConfigParser()
    if os.path.exists(path()) == False: # touch file or config
        open(path(), 'w').close()       # read chokes on missing file
    f = codecs.open(path(), 'r', encoding)
    config.readfp(f, path())
    f.close()
    if value is not None:
        config.set(section, name, value)
    else:
        config.remove_option(section, name)
    f = codecs.open(path(), 'w', encoding)
    config.write(f)
    f.close()

def get_val(name, section="DEFAULT", default=None, encoding=None):
    """Get a value from the per-user config file

    Parameters
    ----------
    name : str
      The name of the value to set.
    section : str
      The section to store the name/value in.
    default :
      The value to return if `name` is not set.
    encoding : str
      The config file's encoding, defaults to :py:data:`default_encoding`.

    Examples
    --------

    >>> get_val("junk") is None
    True
    >>> set_val("junk", "random")
    >>> get_val("junk")
    u'random'
    >>> set_val("junk", None)
    >>> get_val("junk") is None
    True
    """
    if os.path.exists(path()):
        if encoding == None:
            encoding = default_encoding
        config = ConfigParser.ConfigParser()
        f = codecs.open(path(), 'r', encoding)
        config.readfp(f, path())
        f.close()
        try:
            return config.get(section, name)
        except ConfigParser.NoOptionError:
            return default
    else:
        return default

if libbe.TESTING == True:
    suite = doctest.DocTestSuite()



# Copyright (C) 2010-2012 Chris Ball <cjb@laptop.org>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

"""Define the :py:class:`Serve` serving BE Storage over HTTP.

See Also
--------
:py:mod:`libbe.storage.http` : the associated client
"""

import logging
import os.path

import libbe
import libbe.command
import libbe.command.util
import libbe.util.http
import libbe.util.subproc
import libbe.util.wsgi
import libbe.version

if libbe.TESTING:
    import copy
    import doctest
    import StringIO
    import sys
    import unittest
    import wsgiref.validate
    try:
        import cherrypy.test.webtest
        cherrypy_test_webtest = True
    except ImportError:
        cherrypy_test_webtest = None

    import libbe.bugdir
    import libbe.util.wsgi


class ServerApp (libbe.util.wsgi.WSGI_AppObject,
                 libbe.util.wsgi.WSGI_DataObject):
    """WSGI server for a BE Storage instance over HTTP.

    RESTful_ WSGI request handler for serving the
    libbe.storage.http.HTTP backend with GET, POST, and HEAD commands.
    For more information on authentication and REST, see John
    Calcote's `Open Sourcery article`_

    .. _RESTful: http://www.ics.uci.edu/~fielding/pubs/dissertation/rest_arch_style.htm
    .. _Open Sourcery article: http://jcalcote.wordpress.com/2009/08/10/restful-authentication/

    This serves files from a connected storage instance, usually
    a VCS-based repository located on the local machine.

    Notes
    -----

    The GET and HEAD requests are identical except that the HEAD
    request omits the actual content of the file.
    """
    server_version = 'BE-storage-server/' + libbe.version.version()

    def __init__(self, storage=None, notify=False, **kwargs):
        super(ServerApp, self).__init__(
            urls=[
                (r'^add/?', self.add),
                (r'^exists/?', self.exists),
                (r'^remove/?', self.remove),
                (r'^ancestors/?', self.ancestors),
                (r'^children/?', self.children),
                (r'^get/(.+)', self.get),
                (r'^set/(.+)', self.set),
                (r'^commit/?', self.commit),
                (r'^revision-id/?', self.revision_id),
                (r'^changed/?', self.changed),
                (r'^version/?', self.version),
                ],
            **kwargs)
        self.storage = storage
        self.notify = notify

    # handlers
    def add(self, environ, start_response):
        self.check_login(environ)
        data = self.post_data(environ)
        source = 'post'
        id = self.data_get_id(data, source=source)
        parent = self.data_get_string(
            data, 'parent', default=None, source=source)
        directory = self.data_get_boolean(
            data, 'directory', default=False, source=source)
        self.storage.add(id, parent=parent, directory=directory)
        if self.notify:
            self._notify(environ, 'add', id,
                         [('parent', parent), ('directory', directory)])
        return self.ok_response(environ, start_response, None)

    def exists(self, environ, start_response):
        self.check_login(environ)
        data = self.query_data(environ)
        source = 'query'
        id = self.data_get_id(data, source=source)
        revision = self.data_get_string(
            data, 'revision', default=None, source=source)
        content = str(self.storage.exists(id, revision))
        return self.ok_response(environ, start_response, content)

    def remove(self, environ, start_response):
        self.check_login(environ)
        data = self.post_data(environ)
        source = 'post'
        id = self.data_get_id(data, source=source)
        recursive = self.data_get_boolean(
            data, 'recursive', default=False, source=source)
        if recursive == True:
            self.storage.recursive_remove(id)
        else:
            self.storage.remove(id)
        if self.notify:
            self._notify(environ, 'remove', id, [('recursive', recursive)])
        return self.ok_response(environ, start_response, None)

    def ancestors(self, environ, start_response):
        self.check_login(environ)
        data = self.query_data(environ)
        source = 'query'
        id = self.data_get_id(data, source=source)
        revision = self.data_get_string(
            data, 'revision', default=None, source=source)
        content = '\n'.join(self.storage.ancestors(id, revision))+'\n'
        return self.ok_response(environ, start_response, content)

    def children(self, environ, start_response):
        self.check_login(environ)
        data = self.query_data(environ)
        source = 'query'
        id = self.data_get_id(data, default=None, source=source)
        revision = self.data_get_string(
            data, 'revision', default=None, source=source)
        content = '\n'.join(self.storage.children(id, revision))
        return self.ok_response(environ, start_response, content)

    def get(self, environ, start_response):
        self.check_login(environ)
        data = self.query_data(environ)
        source = 'query'
        try:
            id = environ['be-server.url_args'][0]
        except:
            raise libbe.util.wsgi.HandlerError(404, 'Not Found')
        revision = self.data_get_string(
            data, 'revision', default=None, source=source)
        content = self.storage.get(id, revision=revision)
        be_version = self.storage.storage_version(revision)
        return self.ok_response(environ, start_response, content,
                                headers=[('X-BE-Version', be_version)])

    def set(self, environ, start_response):
        self.check_login(environ)
        data = self.post_data(environ)
        try:
            id = environ['be-server.url_args'][0]
        except:
            raise libbe.util.wsgi.HandlerError(404, 'Not Found')
        if not 'value' in data:
            raise libbe.util.wsgi.HandlerError(406, 'Missing query key value')
        value = data['value']
        self.storage.set(id, value)
        if self.notify:
            self._notify(environ, 'set', id, [('value', value)])
        return self.ok_response(environ, start_response, None)

    def commit(self, environ, start_response):
        self.check_login(environ)
        data = self.post_data(environ)
        if not 'summary' in data:
            raise libbe.util.wsgi.HandlerError(
                406, 'Missing query key summary')
        summary = data['summary']
        if not 'body' in data or data['body'] == 'None':
            data['body'] = None
        body = data['body']
        if not 'allow_empty' in data \
                or data['allow_empty'] == 'True':
            allow_empty = True
        else:
            allow_empty = False
        try:
            revision = self.storage.commit(summary, body, allow_empty)
        except libbe.storage.EmptyCommit, e:
            raise libbe.util.wsgi.HandlerError(
                libbe.util.http.HTTP_USER_ERROR, 'EmptyCommit')
        if self.notify:
            self._notify(environ, 'commit', id,
                         [('allow_empty', allow_empty), ('summary', summary),
                          ('body', body)])
        return self.ok_response(environ, start_response, revision)

    def revision_id(self, environ, start_response):
        self.check_login(environ)
        data = self.query_data(environ)
        source = 'query'
        index = int(self.data_get_string(
            data, 'index', default=libbe.util.wsgi.HandlerError,
            source=source))
        content = self.storage.revision_id(index)
        return self.ok_response(environ, start_response, content)

    def changed(self, environ, start_response):
        self.check_login(environ)
        data = self.query_data(environ)
        source = 'query'
        revision = self.data_get_string(
            data, 'revision', default=None, source=source)
        add,mod,rem = self.storage.changed(revision)
        content = '\n\n'.join(['\n'.join(p) for p in (add,mod,rem)])
        return self.ok_response(environ, start_response, content)

    def version(self, environ, start_response):
        self.check_login(environ)
        data = self.query_data(environ)
        source = 'query'
        revision = self.data_get_string(
            data, 'revision', default=None, source=source)
        content = self.storage.storage_version(revision)
        return self.ok_response(environ, start_response, content)

    # handler utility functions
    def check_login(self, environ):
        user = environ.get('be-auth.user', None)
        if user is not None:  # we're running under AuthenticationApp
            if environ['REQUEST_METHOD'] == 'POST':
                if user == 'guest' or self.storage.is_writeable() == False:
                    raise _Unauthorized() # only non-guests allowed to write
            # allow read-only commands for all users

    def _notify(self, environ, command, id, params):
        message = self._format_notification(environ, command, id, params)
        self._submit_notification(message)

    def _format_notification(self, environ, command, id, params):
        key_length = len('command')
        for key,value in params:
            if len(key) > key_length and '\n' not in str(value):
                key_length = len(key)
        key_length += 1
        lines = []
        multi_line_params = []
        for key,value in [('address', environ.get('REMOTE_ADDR', '-')),
                          ('command', command), ('id', id)]+params:
            v = str(value)
            if '\n' in v:
                multi_line_params.append((key,v))
                continue
            lines.append('%*.*s %s' % (key_length, key_length, key+':', v))
        lines.append('')
        for key,value in multi_line_params:
            lines.extend(['=== START %s ===' % key, v,
                          '=== STOP %s ===' % key, ''])
        lines.append('')
        return '\n'.join(lines)

    def _submit_notification(self, message):
        libbe.util.subproc.invoke(self.notify, stdin=message, shell=True)


class ServeStorage (libbe.util.wsgi.ServerCommand):
    """Serve bug directory storage over HTTP.

    This allows you to run local `be` commands interfacing with remote
    data, transmitting file reads/writes/etc. over the network.

    :py:class:`~libbe.command.base.Command` wrapper around
    :py:class:`ServerApp`.
    """

    name = 'serve-storage'

    def _get_app(self, logger, storage, **kwargs):
        return ServerApp(
            logger=logger, storage=storage, notify=kwargs.get('notify', False))

    def _long_help(self):
        return """
Example usage::

    $ be serve-storage

And in another terminal (or after backgrounding the server)::

    $ be --repo http://localhost:8000/ list

If you bind your server to a public interface, take a look at the
``--read-only`` option or the combined ``--ssl --auth FILE``
options so other people can't mess with your repository.  If you do use
authentication, you'll need to send in your username and password::

    $ be --repo http://username:password@localhost:8000/ list
"""


# alias for libbe.command.base.get_command_class()
Serve_storage = ServeStorage


if libbe.TESTING:
    class ServerAppTestCase (libbe.util.wsgi.WSGITestCase):
        def setUp(self):
            super(ServerAppTestCase, self).setUp()
            self.bd = libbe.bugdir.SimpleBugDir(memory=False)
            self.app = ServerApp(self.bd.storage, logger=self.logger)

        def tearDown(self):
            self.bd.cleanup()
            super(ServerAppTestCase, self).tearDown()

        def test_add_get(self):
            try:
                self.getURL(self.app, '/add/', method='GET')
            except libbe.util.wsgi.HandlerError as e:
                self.failUnless(e.code == 404, e)
            else:
                self.fail('GET /add/ did not raise 404')

        def test_add_post(self):
            self.getURL(self.app, '/add/', method='POST',
                        data_dict={'id':'123456', 'parent':'abc123',
                                   'directory':'True'})
            self.failUnless(self.status == '200 OK', self.status)
            self.failUnless(self.response_headers == [],
                            self.response_headers)
            self.failUnless(self.exc_info is None, self.exc_info)
        # Note: other methods tested in libbe.storage.http

        # TODO: integration tests on Serve?

    unitsuite =unittest.TestLoader().loadTestsFromModule(sys.modules[__name__])
    suite = unittest.TestSuite([unitsuite, doctest.DocTestSuite()])



# Copyright (C) 2009-2012 Chris Ball <cjb@laptop.org>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

import glob
import os.path

import libbe
import libbe.command

class Completer (object):
    def __init__(self, options):
        self.options = options
    def __call__(self, bugdirs, fragment=None):
        return [fragment]

def complete_command(command, argument, fragment=None):
    """
    List possible command completions for fragment.

    command argument is not used.
    """
    return list(libbe.command.commands(command_names=True))

def comp_path(fragment=None):
    """List possible path completions for fragment."""
    if fragment == None:
        fragment = '.'
    comps = glob.glob(fragment+'*') + glob.glob(fragment+'/*')
    if len(comps) == 1 and os.path.isdir(comps[0]):
        comps.extend(glob.glob(comps[0]+'/*'))
    return comps

def complete_path(command, argument, fragment=None):
    """List possible path completions for fragment."""
    return comp_path(fragment)

def complete_status(command, argument, fragment=None):
    bd = sorted(command._get_bugdirs().items())[1]
    import libbe.bug
    return libbe.bug.status_values

def complete_severity(command, argument, fragment=None):
    bd = sorted(command._get_bugdirs().items())[1]
    import libbe.bug
    return libbe.bug.severity_values

def assignees(bugdirs):
    ret = set()
    for bugdir in bugdirs.values():
        bugdir.load_all_bugs()
        ret.update(set([bug.assigned for bug in bugdir
                        if bug.assigned != None]))
    return list(ret)

def complete_assigned(command, argument, fragment=None):
    return assignees(command._get_bugdirs())

def complete_extra_strings(command, argument, fragment=None):
    if fragment == None:
        return []
    return [fragment]

def complete_bugdir_id(command, argument, fragment=None):
    bugdirs = command._get_bugdirs()
    return bugdirs.keys()

def complete_bug_id(command, argument, fragment=None):
    return complete_bug_comment_id(command, argument, fragment,
                                   comments=False)

def complete_bug_comment_id(command, argument, fragment=None,
                            active_only=True, comments=True):
    import libbe.bugdir
    import libbe.util.id
    bugdirs = command._get_bugdirs()
    if fragment == None or len(fragment) == 0:
        fragment = '/'
    try:
        p = libbe.util.id.parse_user(bugdirs, fragment)
        matches = None
        root,residual = (fragment, None)
        if not root.endswith('/'):
            root += '/'
    except libbe.util.id.InvalidIDStructure, e:
        return []
    except libbe.util.id.NoIDMatches:
        return []
    except libbe.util.id.MultipleIDMatches, e:
        if e.common == None:
            # choose among bugdirs
            return e.matches
        common = e.common
        matches = e.matches
        root,residual = libbe.util.id.residual(common, fragment)
        p = libbe.util.id.parse_user(bugdirs, e.common)
    bug = None
    if matches == None: # fragment was complete, get a list of children uuids
        if p['type'] == 'bugdir':
            bugdir = bugdirs[p['bugdir']]
            matches = bugdir.uuids()
            common = bugdir.id.user()
        elif p['type'] == 'bug':
            if comments == False:
                return [fragment]
            bugdir = bugdirs[p['bugdir']]
            bug = bugdir.bug_from_uuid(p['bug'])
            matches = bug.uuids()
            common = bug.id.user()
        else:
            assert p['type'] == 'comment', p
            return [fragment]
    if p['type'] == 'bugdir':
        bugdir = bugdirs[p['bugdir']]
        child_fn = bugdir.bug_from_uuid
    elif p['type'] == 'bug':
        if comments == False:
            return[fragment]
        bugdir = bugdirs[p['bugdir']]
        if bug == None:
            bug = bugdir.bug_from_uuid(p['bug'])
        child_fn = bug.comment_from_uuid
    elif p['type'] == 'comment':
        assert matches == None, matches
        return [fragment]
    possible = []
    common += '/'
    for m in matches:
        child = child_fn(m)
        id = child.id.user()
        possible.append(id.replace(common, root))
    return possible

def select_values(string, possible_values, name="unkown"):
    """
    This function allows the user to select values from a list of
    possible values.  The default is to select all the values:

    >>> select_values(None, ['abc', 'def', 'hij'])
    ['abc', 'def', 'hij']

    The user selects values with a comma-separated limit_string.
    Prepending a minus sign to such a list denotes blacklist mode:

    >>> select_values('-abc,hij', ['abc', 'def', 'hij'])
    ['def']

    Without the leading -, the selection is in whitelist mode:

    >>> select_values('abc,hij', ['abc', 'def', 'hij'])
    ['abc', 'hij']

    In either case, appropriate errors are raised if on of the
    user-values is not in the list of possible values.  The name
    parameter lets you make the error message more clear:

    >>> select_values('-xyz,hij', ['abc', 'def', 'hij'], name="foobar")
    Traceback (most recent call last):
      ...
    UserError: Invalid foobar xyz
      ['abc', 'def', 'hij']
    >>> select_values('xyz,hij', ['abc', 'def', 'hij'], name="foobar")
    Traceback (most recent call last):
      ...
    UserError: Invalid foobar xyz
      ['abc', 'def', 'hij']
    """
    possible_values = list(possible_values) # don't alter the original
    if string == None:
        pass
    elif string.startswith('-'):
        blacklisted_values = set(string[1:].split(','))
        for value in blacklisted_values:
            if value not in possible_values:
                raise libbe.command.UserError('Invalid %s %s\n  %s'
                                % (name, value, possible_values))
            possible_values.remove(value)
    else:
        whitelisted_values = string.split(',')
        for value in whitelisted_values:
            if value not in possible_values:
                raise libbe.command.UserError(
                    'Invalid %s %s\n  %s'
                    % (name, value, possible_values))
        possible_values = whitelisted_values
    return possible_values

def bugdir_bug_comment_from_user_id(bugdirs, id):
    p = libbe.util.id.parse_user(bugdirs, id)
    if not p['type'] in ['bugdir', 'bug', 'comment']:
        raise libbe.command.UserError(
            '{0} is a {1} id, not a bugdir, bug, or comment id'.format(
                id, p['type']))
    if p['bugdir'] not in bugdirs:
        raise libbe.command.UserError(
            "{0} doesn't belong to any bugdirs in {1}".format(
                id, sorted(bugdirs.keys())))
    bugdir = bugdirs[p['bugdir']]
    if p['bugdir'] != bugdir.uuid:
        raise libbe.command.UserError(
            "%s doesn't belong to this bugdir (%s)"
            % (id, bugdir.uuid))
    if 'bug' in p:
        bug = bugdir.bug_from_uuid(p['bug'])
        if 'comment' in p:
            comment = bug.comment_from_uuid(p['comment'])
        else:
            comment = bug.comment_root
    else:
        bug = comment = None
    return (bugdir, bug, comment)

def bug_from_uuid(bugdirs, uuid):
    error = None
    for bugdir in bugdirs.values():
        try:
            bug = bugdir.bug_from_uuid(uuid)
        except libbe.bugdir.NoBugMatches as e:
            error = e
        else:
            return bug
    raise error
